
def crystal_input_test(folder_path):
    # Testing of the Crystal_input class

    # folder_path is the path to the folder that contains the test files

    from crystal_io import Crystal_input
    from CRYSTALClear.crystal_io import Crystal_input

    import os

    test_attr = []
    test_result = []

    input_file = os.path.join(folder_path,'mgo.d12')  
    mgo_input = Crystal_input()
    mgo_input.from_file(input_file)

    print('File for input testing: %s' %input_file)

    # GEOM BLOCK

    test_attr.append('geom_block')

    geom_block = ['MGO BULK - GEOMETRY TEST\n',
                    'CRYSTAL\n',
                    '0 0 0\n',
                    '225\n',
                    '4.217\n',
                    '2\n',
                    '12 0.    0.    0.\n',
                    '8 0.5   0.5   0.5\n',
                    'END\n']
    if mgo_input.geom_block == geom_block:
        test_result.append(True)
    else:
        test_result.append(False)

    # BS BLOCK

    test_attr.append('bs_block')

    bs_block = ['12 4\n',
                '0 0 8 2.0 1.0\n',
                ' 68370.0 0.0002226\n',
                ' 9661.0 0.001901\n',
                ' 2041.0 0.011042\n',
                ' 529.6 0.05005\n',
                ' 159.17 0.1690\n',
                ' 54.71 0.36695\n',
                ' 21.236 0.4008\n',
                ' 8.791 0.1487\n',
                '0 1 5 8.0 1.0\n',
                ' 143.7 -0.00671 0.00807\n',
                ' 31.27 -0.07927 0.06401\n',
                ' 9.661 -0.08088 0.2092\n',
                ' 3.726 0.2947 0.3460\n',
                ' 1.598 0.5714 0.3731\n',
                '0 1 1 2.0 1.0\n',
                ' 0.688 1.0 1.0\n',
                '0 1 1 0.0 1.0\n',
                ' 0.28 1.0 1.0\n',
                '8 4\n',
                '0 0 8 2.0 1.0\n',
                ' 8020.0 0.00108\n',
                ' 1338.0 0.00804\n',
                ' 255.4 0.05324\n',
                ' 69.22 0.1681\n',
                ' 23.90 0.3581\n',
                ' 9.264 0.3855\n',
                ' 3.851 0.1468\n',
                ' 1.212 0.0728\n',
                '0 1 4 6.0 1.0\n',
                ' 49.43 -0.00883 0.00958\n',
                ' 10.47 -0.0915 0.0696\n',
                ' 3.235 -0.0402 0.2065\n',
                ' 1.217 0.379 0.347\n',
                '0 1 1 0.0 1.0\n',
                ' 0.4764 1.0 1.0\n',
                '0 1 1 0.0 1.0\n',
                ' 0.1802 1.0 1.0\n',
                '99 0\n',
                'ENDBS\n']

    if mgo_input.bs_block == bs_block:
        test_result.append(True)
    else:
        test_result.append(False)

    # FUNC BLOCK

    test_attr.append('func_block')

    func_block = ['DFT\n', 'B3LYP\n', 'XXLGRID\n', 'ENDDFT\n']

    if mgo_input.func_block == func_block:
        test_result.append(True)
    else:
        test_result.append(False)

    test_attr.append('scf_block')

    # SCF BLOCK

    scf_block = [['TOLINTEG\n', '7 7 7 7 14\n'],
                ['SHRINK\n', '12 24\n'],
                ['MAXCYCLE\n', '200\n'],
                ['FMIXING\n', '70\n'],
                'DIIS\n',
                'ENDSCF\n']
    if mgo_input.scf_block == scf_block:
        test_result.append(True)
    else:
        test_result.append(False)

    # ADD GHOST

    mgo_input = Crystal_input()
    mgo_input.from_file(input_file)
    
    test_attr.append('add_ghost')

    mgo_input.add_ghost([1])
    bs_block = ['GHOSTS\n', '1\n', '1\n', 'ENDBS\n']

    if mgo_input.bs_block[-4:] == bs_block:
        test_result.append(True)
    else:
        test_result.append(False)

    # SP TO OPT

    mgo_input = Crystal_input()
    mgo_input.from_file(input_file)
    
    test_attr.append('sp_to_opt')

    mgo_input.sp_to_opt()
    geom_block = ['OPTGEOM\n', 'END\n', 'END\n']

    if mgo_input.geom_block[-3:] == geom_block:
        test_result.append(True)
    else:
        test_result.append(False)
    
    # OPT TO SP

    test_attr.append('opt_to_sp')

    mgo_input.opt_to_sp()
    geom_block = ['MGO BULK - GEOMETRY TEST\n',
                    'CRYSTAL\n',
                    '0 0 0\n',
                    '225\n',
                    '4.217\n',
                    '2\n',
                    '12 0.    0.    0.\n',
                    '8 0.5   0.5   0.5\n',
                    'END\n']

    if mgo_input.geom_block == geom_block:
        test_result.append(True)
    else:
        test_result.append(False)


    return [test_attr, test_result]


def crystal_output_test(folder_path):
    # Testing of the Crystal_output class

    # folder_path is the path to the folder that contains the test files
    from crystal_io import Crystal_output
    from CRYSTALClear.crystal_io import Crystal_output

    import numpy as np
    import os

    test_attr = []
    test_result = []

    output_file = os.path.join(folder_path,'mgo.out')  
    output_opt_file = os.path.join(folder_path,'mgo_optgeom.out')  
    mgo_output = Crystal_output().read_cry_output(output_file)
    mgo_opt_output = Crystal_output().read_cry_output(output_opt_file)

    print('Files for input testing: %s and %s' %(output_file,output_opt_file))

    # CONVERGED

    test_attr.append('converged')

    if mgo_output.converged == True:
        test_result.append(True)
    else:
        test_result.append(False)

    # GET DIMENSIONALITY

    test_attr.append('get_dimensionality')

    if mgo_output.get_dimensionality() == 3:
        test_result.append(True)
    else:
        test_result.append(False)

    # GET DIMENSIONALITY

    test_attr.append('get_final_energy')

    if abs(mgo_output.get_final_energy() - (-7495.341792877063)) < 10E-6:
        test_result.append(True)
    else:
        test_result.append(False)

    
    # GET SCF CONVERGENCE

    test_attr.append('get_scf_convergence')
    
    get_scf_convergence_0 = np.array([-7476.83111864, -7489.9832359 , -7494.92298163, -7495.34019539,
       -7495.34166646, -7495.34179102, -7495.34179278, -7495.34179288])
    get_scf_convergence_1 = np.array([-7.48313500e+03, -1.31431062e+01, -4.95247480e+00, -4.16334420e-01,
       -1.47213674e-03, -1.24628212e-04, -1.75513530e-06, -1.02586978e-07])
    
    if abs(np.sum(mgo_output.get_scf_convergence()[0] - get_scf_convergence_0) + \
        np.sum(mgo_output.get_scf_convergence()[1] - get_scf_convergence_1)) < 10E-6:
        test_result.append(True)
    else:
        test_result.append(False)

    # GET NUM CYCLES

    test_attr.append('get_num_cycles')

    if mgo_output.get_num_cycles() == 7:
        test_result.append(True)
    else:
        test_result.append(False)

    # GET FERMI ENERGY

    test_attr.append('get_fermi_energy')

    fermi_energy = -4.13671240282

    if abs(mgo_output.get_fermi_energy() - fermi_energy) < 10E-6:
        test_result.append(True)
    else:
        test_result.append(False)

    # GET PRIMITIVE LATTICE

    test_attr.append('get_primitive_lattice')

    primitive_lattice = np.array([[0.    , 2.1085, 2.1085],
                                    [2.1085, 0.    , 2.1085],
                                    [2.1085, 2.1085, 0.    ]])
    
    if np.all(mgo_output.get_primitive_lattice() == primitive_lattice):
        test_result.append(True)
    else:
        test_result.append(False)

    # GET reciprocal LATTICE

    test_attr.append('get_reciprocal_lattice')

    reciprocal_lattice = np.array([[-1.48996571,  1.48996571,  1.48996571],
                                    [ 1.48996571, -1.48996571,  1.48996571],
                                    [ 1.48996571,  1.48996571, -1.48996571]])
    returned_reciprocal_lattice = np.round(np.array(mgo_output.get_reciprocal_lattice()),8)
    
    if np.all(returned_reciprocal_lattice == reciprocal_lattice):
        test_result.append(True)
    else:
        test_result.append(False)
    
    # GET BAND GAP

    test_attr.append('get_band_gap')

    if abs(mgo_output.get_band_gap() - 7.1237) < 10E-6:
        test_result.append(True)
    else:
        test_result.append(False)
    
    # GET LAST GEOM

    test_attr.append('get_last_geom')

    get_lat_geom_0 = np.array([[0.    , 2.1085, 2.1085],
                                [2.1085, 0.    , 2.1085],
                                [2.1085, 2.1085, 0.    ]])

    get_lat_geom_1 = np.array([12,  8])

    get_lat_geom_2 = np.array([[0.    , 0.    , 0.    ],
                                [2.1085, 2.1085, 2.1085]])

    if np.all(mgo_output.get_last_geom(write_gui_file=False)[0] == get_lat_geom_0) and \
        np.all(mgo_output.get_last_geom(write_gui_file=False)[1] == get_lat_geom_1) and \
        np.all(mgo_output.get_last_geom(write_gui_file=False)[2] == get_lat_geom_2):
        test_result.append(True)
    else:
        test_result.append(False)

    # GET SYMMOPS

    test_attr.append('get_symm_ops')

    len_symm_ops = 48
    symm_ops_0 = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]

    if len(mgo_output.get_symm_ops()) == len_symm_ops and \
        mgo_output.get_symm_ops()[0].tolist() == symm_ops_0:
        test_result.append(True)
    else:
        test_result.append(False)

    # GET FORCES

    test_attr.append('get_forces')

    forces_0 = np.array([[-0.00148273,  0.00148273,  0.00148273],
       [ 0.00148273, -0.00148273,  0.00148273],
       [ 0.00148273,  0.00148273, -0.00148273]])
    forces_1 = np.round(np.array([[-4.18889278e-16, -4.11372640e-16, -4.14981546e-16],
       [ 4.14263349e-16,  4.13107364e-16,  4.14981546e-16]]),8)

    returned_forces_0 = np.round(np.array(mgo_opt_output.get_forces()[0]),8)
    returned_forces_1 = np.round(np.array(mgo_opt_output.get_forces()[1]),8)

    if np.all(returned_forces_0 == forces_0) and np.all(returned_forces_1 == forces_1):
        test_result.append(True)
    else:
        test_result.append(False)

    # GET MULLIKEN CHARGES

    test_attr.append('get_mulliken_charges')

    mulliken_charges = [10.129, 9.871]

    if np.all(mgo_opt_output.get_mulliken_charges() == mulliken_charges):
        test_result.append(True)
    else:
        test_result.append(False)

    return [test_attr, test_result]

def convert_test(folder_path):
    # Testing of the functions contained in convert_test

    # folder_path is the path to the folder that contains the test files

    import os
    import numpy as np
    from crystal_io import Crystal_output, Crystal_gui
    from CRYSTALClear.crystal_io import Crystal_output, Crystal_gui
    from convert import cry_out2pmg, cry_gui2pmg
    from CRYSTALClear.convert import cry_out2pmg, cry_gui2pmg

    test_attr = []
    test_result = []

    output_opt_file = os.path.join(folder_path,'mgo_optgeom.out')  
    gui_file = os.path.join(folder_path,'mgo_optgeom.gui')
    mgo_opt_output = Crystal_output().read_cry_output(output_opt_file)

    # CRY_OUT2PMG

    test_attr.append('cry_out2pmg')

    pmg_obj = cry_out2pmg(mgo_opt_output)
    
    lattice_matrix = np.array([[0.        , 2.12011001, 2.12011001],
       [2.12011001, 0.        , 2.12011001],
       [2.12011001, 2.12011001, 0.        ]])
    cart_coord = np.array([[0.        , 0.        , 0.        ],
       [-2.12011001, -2.12011001, -2.12011001]])

    if np.all(np.round(pmg_obj.lattice.matrix,8) == lattice_matrix) and \
        np.all(np.round(pmg_obj.cart_coords,8) == cart_coord):
        test_result.append(True)
    else:
        test_result.append(False)

    # CRY_GUI2PMG

    test_attr.append('cry_gui2pmg')
    mgo_gui = Crystal_gui()
    mgo_gui.read_gui(gui_file)
    pmg_obj = cry_gui2pmg(mgo_gui)
    
    lattice_matrix = np.round(np.array([[-1.29819297e-16,  2.12011001e+00,  2.12011001e+00],
       [ 2.12011001e+00,  0.00000000e+00,  2.12011001e+00],
       [ 2.12011001e+00,  2.12011001e+00,  2.59638594e-16]]),8)
    
    cart_coord = np.round(np.array([[0.     , 0.     , 0.     ],
       [2.12011, 2.12011, 2.12011]]),8)

    if np.all(np.round(pmg_obj.lattice.matrix,8) == lattice_matrix) and \
        np.all(np.round(pmg_obj.cart_coords,8) == cart_coord):
        test_result.append(True)
    else:
        test_result.append(False)

    return [test_attr, test_result]


def test_all(folder_path):
    # Testing of all the above classes and functions

    # folder_path is the path to the folder that contains the test files
       
    print('***Testing of the crystal_io module:***\n')

    print('**Testing of the Crystal_input class:**\n')
    
    crystal_input_test_out = crystal_input_test(folder_path)
    
    print('')
    
    for i in range(len(crystal_input_test_out[0])):
        print(crystal_input_test_out[0][i],crystal_input_test_out[1][i])

    print('\n**Testing of the Crystal_output class:**\n')

    crystal_output_test_out = crystal_output_test(folder_path)
    
    print('')

    for i in range(len(crystal_output_test_out[0])):
        print(crystal_output_test_out[0][i],crystal_output_test_out[1][i])

    print('\n***Testing of the convert module:***\n')

    convert_test_out = convert_test(folder_path)
    
    for i in range(len(convert_test_out[0])):
        print(convert_test_out[0][i],convert_test_out[1][i])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes and methods of keywords used in 'crystal' input file (d12).
"""
from CRYSTALClear.base.inputbase import BlockBASE

class Crystal_inputBASE(BlockBASE):
    """
    The base class of Crystal_input class
    """

    def __init__(self):
        # Initialize the object to empty values
        self._block_bg = None  # Avoid empty bg and ed lines
        self._block_ed = None
        self._block_data = ''
        self._block_attr = ['geom', 'basisset', 'scf']

    @property
    def geom(self):
        if not hasattr(self, '_block_geom'):
            self.set_geom()
        return self._block_geom

    def set_geom(self, obj=None):
        """
        Geom subblock

        Args:
            obj (Geom | str): A block object of 'GEOM' submodule. Or a string
                in CRYSTAL d12 format.
        """
        from CRYSTALClear.base.crysd12 import Geom

        self._block_geom = Geom()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_geom.clean_block()
            else:
                dimen_list = ['CRYSTAL\n', 'SLAB\n', 'POLYMER\n', 'MOLECULE\n',
                              'HELIX\n', 'EXTERNAL\n', 'DLVINPUT\n']
                title = obj.split('\n')[0]
                if title in dimen_list: # No title line
                    self._block_geom.analyze_text(obj)
                else: # Have title line
                    self._block_geom.title(title)
                    self._block_geom.analyze_text(obj)
        else:
            self._block_geom = obj

    @property
    def basisset(self):
        if not hasattr(self, '_block_basisset'):
            self.set_basisset()
        return self._block_basisset

    def set_basisset(self, obj=None):
        """
        Basis set subblock

        Args:
            obj (BasisSet | str): A block object of basis set submodule. Or a
                string in CRYSTAL d12 format
        """
        from CRYSTALClear.base.crysd12 import BasisSet

        self._block_basisset = BasisSet()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '':# Clean data
                self._block_basisset.clean_block()
            else:
                self._block_basisset.analyze_text(obj)
        else:
            self._block_basisset = obj

    @property
    def scf(self):
        if not hasattr(self, '_block_scf'):
            self.set_scf()
        return self._block_scf

    def set_scf(self, obj=None):
        """
        SCF subblock

        Args:
            obj (SCF | str): A block object of SCF submodule. Or a string in
                CRYSTAL d12 format
        """
        from CRYSTALClear.base.crysd12 import SCF

        self._block_scf = SCF()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_scf.clean_block()
            else:
                self._block_scf.analyze_text(obj)
        else:
            self._block_scf = obj

    def from_file(self, file):
        """
        Generate a CrystalInputBASE obj from a d12 file. A 'complete' d12 file
        with geometry, basis set and SCF blocks is suggested.
        """
        import re

        inp = open(file, 'r')
        data = inp.read()
        inp.close()
        data_lines = data.strip().split('\n')

        self.set_geom()
        self.set_basisset()
        self.set_scf()

        # Divide data into 3 blocks
        text = ['', '', '']
        # Case 1: No BASISSET keyword
        if 'BASISSET' not in data:
            geom_end = 1
            bs_end = 2
            end_counter = 0
            block_counter = 0
            subblock_key = ['OPTGEOM', 'FREQCALC',
                            'ANHARM', 'CPHF', 'CPKS', 'ELASTCON', 'EOS']
            for d in data_lines:
                d = d.strip()
                text[block_counter] += d + '\n'
                if d in subblock_key:
                    geom_end += 1
                    bs_end += 1
                elif 'END' in d:
                    end_counter += 1

                if end_counter == geom_end:
                    geom_end = -1
                    block_counter += 1
                elif end_counter == bs_end:
                    bs_end = -1
                    block_counter += 1
        # Case 2: BASISSET keyword
        else:
            self.basisset._block_ed = None
            self.geom._block_ed = None
            block_counter = 0
            for d in data_lines:
                d = d.strip()
                if d == 'BASISSET':
                    block_counter += 1
                    text[block_counter] += d + '\n'
                elif d in self.scf._block_key and block_counter == 1:  # Avoid same keywords. e.g. TOLDEE
                    block_counter += 1
                    text[block_counter] += d + '\n'
                else:
                    text[block_counter] += d + '\n'

        # Title line
        self.geom.title(data_lines[0])
        # The 0D geometry keyword 'MOLECULE' and the molecular crystal option 'MOLECULE'
        if 'MOLECULE' not in data_lines[1] and 'MOLECULE' in text[0]:
            text[0] = re.sub('MOLECULE', 'MOLEISO', text[0])

        self.geom.analyze_text(text[0])
        if 'BASISSET' not in data:
            self.basisset.from_string(text[1])
        else:
            self.basisset.analyze_text(text[1])
        self.scf.analyze_text(text[2])

        return self

    def to_file(self, file):
        """
        Write data to a file
        """
        out = open(file, 'w')
        # If objects are set separately, BASISSET might coincide with 'ENDGEOM'
        if 'BASISSET' in self.data and 'ENDGEOM' in self.data:
            self.geom._block_ed = ''
        out.write('%s' % self.data)
        out.close()


class Geom(BlockBASE):
    """
    Geometry block object
    """

    def __init__(self):
        self._block_bg = 'Generated by CRYSTALClear\n'  # Set title as bg label
        self._block_ed = 'ENDGEOM\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'CRYSTAL'   : '_basegeom',
            'SLAB'      : '_basegeom',
            'POLYMER'   : '_basegeom',
            'HELIX'     : '_basegeom',
            'MOLECULE'  : '_basegeom',
            'EXTERNAL'  : '_basegeom',
            'DLVINPUT'  : '_basegeom',
            'SUPERCEL'  : '_sp_matrix',
            'SUPERCELL' : '_sp_matrix',
            'SUPERCON'  : '_sp_matrix',
            'SCELCONF'  : '_sp_matrix',
            'SCELPHONO' : '_sp_matrix',
            'ATOMBSSE'  : '_atombsse',
            'ATOMDISP'  : '_atomdisp',
            'ATOMINSE'  : '_atominse',
            'ATOMORDE'  : '_atomorde',
            'ATOMREMO'  : '_atomremo',
            'MOLEISO'   : '_moleiso', # The original 'MOLECULE' option to isolate molecules
            'EXTPRT'    : '_extprt',
            'CIFPRT'    : '_cifprt',
            'CIFPRTSYM' : '_cifprtsym',
            'COORPRT'   : '_coorprt',
            'TESTGEOM'  : '_testgeom',
            'OPTGEOM'   : 'optgeom',  # Sub-block properties must be named without the initial underscore
            'FREQCALC'  : 'freqcalc',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    @property
    def data(self):
        """
        Settings in all the attributes are summarized here. Covers the data
        property in BlockBASE, to address the ambiguity of 'MOLECULE' keywords.
        """
        import re

        self.update_block()
        text = ''
        for i in [self._block_bg, self._block_data, self._block_ed]:
            if i == None:
                continue
            text += i

        return re.sub('MOLEISO', 'MOLECULE', text)

    def title(self, title='Generated by CRYSTALClear'):
        self._block_bg = '{}\n'.format(title)

    def crystal(self, IGR=None, latt=[], atom=[], IFLAG=0, IFHR=0, IFSO=0, origin=[]):
        """
        Define 'CRYSTAL' structure

        Args:
            sg (int): Space group number. Parameter IGR in the manual
            latt (list): Minimal set of crystallographic cell parameters
            atom (list): Natom \* 4 list of conventional atomic number and 3D 
                fractional coordinates.
            IFLAG (int): See the manual
            IFHR (int): See the manual
            IFSO (int): See the manual
            origin (list): *IFSO > 1* See the manual
        """
        if IGR == None:  # No entry. Return keyword
            self._basegeom = super(Geom, self).assign_keyword('CRYSTAL', [])
            return
        elif IGR == '':  # Clean data
            self._basegeom = super(Geom, self).assign_keyword('CRYSTAL', [], '')
            return

        if IFSO <= 1:
            shape = [3, 1]
            value = ['{:<3d}'.format(int(IFLAG)), '{:<3d}'.format(int(IFHR)), '{:<3d}'.format(int(IFSO)),
                     '{:<3d}'.format(int(IGR))]
        else:
            shape = [3, 3, 1]
            value = ['{:<3d}'.format(int(IFLAG)), '{:<3d}'.format(int(IFHR)), '{:<3d}'.format(int(IFSO)),
                     '{:<12.6f}'.format(origin[0]), '{:<12.6f}'.format(origin[1]), '{:<12.6f}'.format(origin[2]),
                     '{:<3d}'.format(int(IGR))]

        shape += [len(latt), ]
        value += ['{:<12.6f}'.format(i) for i in latt]

        # Format atoms - frac 12 digits, cart 16 digits
        atom = [['{:<4d}'.format(int(a[0])), '{0: 12.8f}'.format(a[1]),
                 '{0: 12.8f}'.format(a[2]), '{0: 12.8f}'.format(a[3])] for a in atom]
        atominput = super(Geom, self).set_list(len(atom), atom)
        shape += atominput[0]
        value += atominput[1]

        self._basegeom = super(Geom, self).assign_keyword('CRYSTAL', shape, value)

    def slab(self, IGR=None, latt=[], atom=[]):
        """
        Define 'SLAB' structure
        """
        if IGR == None:  # No entry. Return keyword
            self._basegeom = super(Geom, self).assign_keyword('SLAB', [])
            return
        elif IGR == '':  # Clean data
            self._basegeom = super(Geom, self).assign_keyword('SLAB', [], '')
            return

        shape = [1, ]
        value = ['{:<3d}'.format(int(IGR)), ]

        shape += [len(latt), ]
        value += ['{:<12.6f}'.format(i) for i in latt]

        # Format atoms - frac 12 digits, cart 16 digits
        atom = [['{:<4d}'.format(int(a[0])), '{0: 12.8f}'.format(a[1]),
                 '{0: 12.8f}'.format(a[2]), '{0: 16.8f}'.format(a[3])] for a in atom]
        atominput = super(Geom, self).set_list(len(atom), atom)
        shape += atominput[0]
        value += atominput[1]

        self._basegeom = super(Geom, self).assign_keyword('SLAB', shape, value)

    def polymer(self, IGR=None, latt=[], atom=[]):
        """
        Define 'POLYMER' structure
        """
        if IGR == None:  # No entry. Return keyword
            self._basegeom = super(Geom, self).assign_keyword('POLYMER', [])
            return
        elif IGR == '':  # Clean data
            self._basegeom = super(Geom, self).assign_keyword('POLYMER', [], '')
            return

        shape = [1, ]
        value = ['{:<3d}'.format(int(IGR)), ]

        shape += [len(latt), ]
        value += ['{:<12.6f}'.format(i) for i in latt]

        # Format atoms - frac 12 digits, cart 16 digits
        atom = [['{:<4d}'.format(int(a[0])), '{0: 12.8f}'.format(a[1]),
                 '{0: 16.8f}'.format(a[2]), '{0: 16.8f}'.format(a[3])] for a in atom]
        atominput = super(Geom, self).set_list(len(atom), atom)
        shape += atominput[0]
        value += atominput[1]

        self._basegeom = super(Geom, self).assign_keyword('POLYMER', shape, value)

    def helix(self, N1=None, N2=0, latt=[], atom=[]):
        """
        Define 'HELIX' structure

        Args:
            N1 (int): See the manual
            N2 (int): See the manual
        """
        if N1 == None:  # No entry. Return keyword
            self._basegeom = super(Geom, self).assign_keyword('HELIX', [])
            return
        elif N1 == '':  # Clean data
            self._basegeom = super(Geom, self).assign_keyword('HELIX', [], '')
            return

        shape = [2, ]
        value = ['{:<3d}'.format(int(N1)), '{:<3d}'.format(int(N2)), ]

        shape += [len(latt), ]
        value += ['{:<12.6f}'.format(i) for i in latt]

        # Format atoms - frac 12 digits, cart 16 digits
        atom = [['{:<4d}'.format(int(a[0])), '{0: 12.8f}'.format(a[1]),
                 '{0: 16.8f}'.format(a[2]), '{0: 16.8f}'.format(a[3])] for a in atom]
        atominput = super(Geom, self).set_list(len(atom), atom)
        shape += atominput[0]
        value += atominput[1]

        self._basegeom = super(Geom, self).assign_keyword('HELIX', shape, value)

    def molecule(self, IGR=None, atom=[]):
        """
        Define 'MOLECULE' structure
        """
        import warnings

        # The 0D geometry keyword 'MOLECULE' and the molecular crystal option 'MOLECULE'
        if hasattr(self, '_basegeom'):
            warnings.warn("Geometry definition exists. To launch the MOLECULE option that isolates molecules from lattice, use the 'self.moleiso' method.", stacklevel=2)

        if IGR == None:  # No entry. Return keyword
            self._basegeom = super(Geom, self).assign_keyword('MOLECULE', [])
            return
        elif IGR == '':  # Clean data
            self._basegeom = super(Geom, self).assign_keyword('MOLECULE', [], '')
            return

        shape = [1, ]
        value = ['{:<3d}'.format(int(IGR)), ]

        # Format atoms - frac 12 digits, cart 16 digits
        atom = [['{:<4d}'.format(int(a[0])), '{0: 16.8f}'.format(a[1]),
                 '{0: 16.8f}'.format(a[2]), '{0: 16.8f}'.format(a[3])] for a in atom]
        atominput = super(Geom, self).set_list(len(atom), atom)
        shape += atominput[0]
        value += atominput[1]

        self._basegeom = super(Geom, self).assign_keyword('MOLECULE', shape, value)

    def external(self, key='EXTERNAL'):
        """
        Define 'EXTERNAL' structure
        """
        self._basegeom = super(Geom, self).assign_keyword(key, [])

    def dlvinput(self, key='DLVINPUT'):
        """
        Define 'DLVINPUT' structure
        """
        self._basegeom = super(Geom, self).assign_keyword(key, [])

    def supercel(self, mx=None):
        """
        Supercell by 'SUPERCEL' keyword

        Args:
            mx (array | list | str): ndimen \* ndimen matrix, [] or ''
        """
        shape, value = super(Geom, self).set_matrix(mx)
        self._sp_matrix = super(Geom, self).assign_keyword('SUPERCEL', shape, value)

    def supercon(self, mx=None):
        """
        Supercell by 'SUPERCON' keyword
        """
        shape, value = super(Geom, self).set_matrix(mx)
        self._sp_matrix = super(Geom, self).assign_keyword('SUPERCON', shape, value)

    def scelconf(self, mx=None):
        """
        Supercell by 'SCELCONF' keyword
        """
        shape, value = super(Geom, self).set_matrix(mx)
        self._sp_matrix = super(Geom, self).assign_keyword('SCELCONF', shape, value)

    def scelphono(self, mx=None):
        """
        Supercell by 'SCELPHONO' keyword
        """
        shape, value = super(Geom, self).set_matrix(mx)
        self._sp_matrix = super(Geom, self).assign_keyword('SCELPHONO', shape, value)

    def atombsse(self, IAT=None, NSTAR=None, RMAX=None):
        self._atombsse = super(Geom, self).assign_keyword('ATOMBSSE', [3, ], [IAT, NSTAR, RMAX])

    def atomdisp(self, NDISP=None, atom=[]):
        """
        ATOMDISP keyword

        Args:
            NDISP (int): See manual
            atom (list): NDISP\*4 list. Including LB, DX, DY, DZ
        """
        shape, value = super(Geom, self).set_list(NDISP, atom)
        self._atomdisp = super(Geom, self).assign_keyword('ATOMDISP', shape, value)

    def atominse(self, NINS=None, atom=[]):
        """
        ATOMINSE keyword

        Args:
            NINS (int): See manual
            atom (list): NINS\*4 list. Including NA, X, Y, Z
        """
        shape, value = super(Geom, self).set_list(NINS, atom)
        self._atominse = super(Geom, self).assign_keyword('ATOMINSE', shape, value)

    def atomorde(self, key='ATOMORDE'):
        self._atomorde = super(Geom, self).assign_keyword(key, [])

    def atomremo(self, NL=None, atom=[]):
        """
        ATOMREMO keyword

        Args:
            NL (int): See manual
            atom (list): NL\*1 list. Including LB
        """
        shape, value = super(Geom, self).set_list(NL, atom)
        self._atomremo = super(Geom, self).assign_keyword('ATOMREMO', shape, value)

    def moleiso(self, NMOL=None, atom=[]):
        """
        .. note::

            Corresponds to the **option** MOLECULE to isolate molecules from
            lattice. Only the method's name is changed to avoid ambiguity with
            the one to set 0D geometry.

        Args:
            NMOL (int)
            atom (list[list[int]]): NMOL\*4 list. See CRYSTAL manual.
        """
        shape, value = super(Geom, self).set_list(NMOL, atom)
        self._moleiso = super(Geom, self).assign_keyword('MOLECULE', shape, value)

    def extprt(self, key='EXTPRT'):
        self._extprt = super(Geom, self).assign_keyword(key, [])

    def cifprt(self, key='CIFPRT'):
        self._cifprt = super(Geom, self).assign_keyword(key, [])

    def cifprtsym(self, key='CIFPRTSYM'):
        self._cifprtsym = super(Geom, self).assign_keyword(key, [])

    def coorprt(self, key='COORPRT'):
        self._coorprt = super(Geom, self).assign_keyword(key, [])

    def testgeom(self, key='TESTGEOM'):
        conflict = ['_block_optgeom', '_block_freqcalc', '_testgeom']
        super(Geom, self).clean_conflict('_testgeom', conflict)

        self._testgeom = super(Geom, self).assign_keyword(key, [])

    @property
    def optgeom(self):
        """
        Subblock object OPTGEOM
        """
        if not hasattr(self, '_block_optgeom'):
            self.set_optgeom()
        return self._block_optgeom

    def set_optgeom(self, obj=None):
        """
        Optgeom subblock

        Args:
            obj (Optgeom | str): A block object of 'OPTGEOM' submodule. Or a
                string in CRYSTAL d12 format
        """
        from CRYSTALClear.base.crysd12 import Optgeom

        conflict = ['_block_optgeom', '_block_freqcalc', '_testgeom']
        super(Geom, self).clean_conflict('_block_optgeom', conflict)

        self._block_optgeom = Optgeom()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_optgeom.clean_block()
            else:
                self._block_optgeom.analyze_text(obj)
        else:
            self._block_optgeom = obj

    @property
    def freqcalc(self):
        """
        Subblock object FREQCALC
        """
        if not hasattr(self, '_block_freqcalc'):
            self.set_freqcalc()
        return self._block_freqcalc

    def set_freqcalc(self, obj=None):
        """
        Freqcalc subblock

        Args:
            obj (Freqcalc | str): A block object of 'FREQCALC' submodule. Or a
                string in CRYSTAL d12 format
        """
        from CRYSTALClear.base.crysd12 import Freqcalc

        conflict = ['_block_optgeom', '_block_freqcalc', '_testgeom']
        super(Geom, self).clean_conflict('_block_freqcalc', conflict)

        self._block_freqcalc = Freqcalc()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_freqcalc.clean_block()
            else:
                self._block_freqcalc.analyze_text(obj)
        else:
            self._block_freqcalc = obj


class Optgeom(BlockBASE):
    """
    OPTGEOM block object
    """

    def __init__(self):
        self._block_bg = 'OPTGEOM\n'
        self._block_ed = 'ENDOPT\n'
        self._block_data = ''
        self._block_dict = {
            'FULLOPTG'    : '_opttype',
            'CELLONLY'    : '_opttype',
            'INTREDUN'    : '_opttype',
            'ITATOCEL'    : '_opttype',
            'CVOLOPT'     : '_opttype',
            'HESSIDEN'    : '_opthess',
            'HESSMOD1'    : '_opthess',
            'HESSMOD2'    : '_opthess',
            'HESSNUM'     : '_opthess',
            'TOLDEG'      : '_toldeg',
            'TOLDEX'      : '_toldex',
            'TOLDEE'      : '_toldee',
            'MAXCYCLE'    : '_maxcycle',
            'FRAGMENT'    : '_fragment',
            'RESTART'     : '_restart',
            'FINALRUN'    : '_finalrun',
            'EXTPRESS'    : '_extpress',
            'ALLOWTRUSTR' : '_usetrustr',
            'NOTRUSTR'    : '_usetrustr',
            'MAXTRADIUS'  : '_maxtradius',
            'TRUSTRADIUS' : '_trustradius',
            'ONELOG'      : '_printonelog',
            'NOXYZ'       : '_printxyz',
            'NOSYMMOPS'   : '_printsymmops',
            'PRINTFORCES' : '_printforces',
            'PRINTHESS'   : '_printhess',
            'PRINTOPT'    : '_printopt',
            'PRINT'       : '_print',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    def fulloptg(self, key='FULLOPTG'):
        self._opttype = super(Optgeom, self).assign_keyword(key, [])

    def cellonly(self, key='CELLONLY'):
        self._opttype = super(Optgeom, self).assign_keyword(key, [])

    def intredun(self, key='INTREDUN'):
        self._opttype = super(Optgeom, self).assign_keyword(key, [])

    def itatocel(self, key='ITATOCEL'):
        self._opttype = super(Optgeom, self).assign_keyword(key, [])

    def cvolopt(self, key='CVOLOPT'):
        self._opttype = super(Optgeom, self).assign_keyword(key, [])

    def hessiden(self, key='HESSIDEN'):
        self._opthess = super(Optgeom, self).assign_keyword(key, [])

    def hessmod1(self, key='HESSMOD1'):
        self._opthess = super(Optgeom, self).assign_keyword(key, [])

    def hessmod2(self, key='HESSMOD2'):
        self._opthess = super(Optgeom, self).assign_keyword(key, [])

    def hessnum(self, key='HESSNUM'):
        self._opthess = super(Optgeom, self).assign_keyword(key, [])

    def toldeg(self, TG=None):
        self._toldeg = super(Optgeom, self).assign_keyword('TOLDEG', [1, ], TG)

    def toldex(self, TX=None):
        self._toldex = super(Optgeom, self).assign_keyword('TOLDEX', [1, ], TX)

    def toldee(self, IG=None):
        self._toldee = super(Optgeom, self).assign_keyword('TOLDEE', [1, ], IG)

    def maxcycle(self, MAX=None):
        self._maxcycle = super(Optgeom, self).assign_keyword('MAXCYCLE', [1, ], MAX)

    def fragment(self, NL=None, LB=[]):
        """
        Args:
            NL (int | str): Number of atoms. See manual. Or ''
            LB (list[int]): Label of atoms. See manual
        """
        shape, value = super(Optgeom, self).set_list(NL, LB)
        self._fragment = super(Optgeom, self).assign_keyword('FRAGMENT', shape, value)

    def restart(self, key='RESTART'):
        self._restart = super(Optgeom, self).assign_keyword(key, [])

    def finalrun(self, ICODE=None):
        self._finalrun = super(Optgeom, self).assign_keyword('FINALRUN', [1, ], ICODE)

    def extpress(self, pres=None):
        self._extpress = super(Optgeom, self).assign_keyword('EXTPRESS', [1, ], pres)

    def allowtrustr(self, key='ALLOWTRUSTR'):
        self._usetrustr = super(Optgeom, self).assign_keyword(key, [])

    def notrustr(self, key='NOTRUSTR'):
        self._usetrustr = super(Optgeom, self).assign_keyword(key, [])

    def maxtradius(self, TRMAX=None):
        import warnings

        if hasattr(self, '_usetrustr'):
            if self._usetrustr == 'NOTRUSTR\n':
                warnings.warn("The pre-set 'NOTRUSTR' keyword will be removed.", stacklevel=2)
                self.notrustr('')
        self._maxtradius = super(Optgeom, self).assign_keyword('MAXTRADIUS', [1, ], TRMAX)

    def trustradius(self, TRADIUS=None):
        import warnings

        if hasattr(self, '_usetrustr'):
            if self._usetrustr == 'NOTRUSTR\n':
                warnings.warn("The pre-set 'NOTRUSTR' keyword will be removed.", stacklevel=2)
                self.notrustr('')
        self._trustradius = super(Optgeom, self).assign_keyword('TRUSTRADIUS', [1, ], TRADIUS)

    def onelog(self, key='ONELOG'):
        self._printonelog = super(Optgeom, self).assign_keyword(key, [])

    def noxyz(self, key='NOXYZ'):
        self._printxyz = super(Optgeom, self).assign_keyword(key, [])

    def nosymmops(self, key='NOSYMMOPS'):
        self._printsymmops = super(Optgeom, self).assign_keyword(key, [])

    def printforces(self, key='PRINTFORCES'):
        self._printforces = super(Optgeom, self).assign_keyword(key, [])

    def printhess(self, key='PRINTHESS'):
        self._printhess = super(Optgeom, self).assign_keyword(key, [])

    def printopt(self, key='PRINTOPT'):
        self._printopt = super(Optgeom, self).assign_keyword(key, [])

    def print(self, key='PRINT'):
        self._print = super(Optgeom, self).assign_keyword(key, [])


class Freqcalc(BlockBASE):
    """
    FREQCALC block object
    """

    def __init__(self):
        self._block_bg = 'FREQCALC\n'
        self._block_ed = 'ENDFREQ\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'NOOPTGEOM'  : '_nooptgeom',  # A tailored sub-block
            'PREOPTGEOM' : 'optgeom',  # A tailored sub-block
            'DISPERSION' : '_dispersion',
            'BANDS'      : '_bands',
            'NUMDERIV'   : '_numderiv',
            'STEPSIZE'   : '_stepsize',
            'RESTART'    : '_restart',
            'MODES'      : '_modes',
            'NOMODES'    : '_modes',
            'PRESSURE'   : '_pressure',
            'TEMPERAT'   : '_temperat',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    def nooptgeom(self, key='NOOPTGEOM'):
        conflict = ['_block_optgeom', '_nooptgeom']
        super(Freqcalc, self).clean_conflict('_nooptgeom', conflict)

        self._nooptgeom = super(Freqcalc, self).assign_keyword(key, [])

    @property
    def optgeom(self):
        try:
            return self._block_optgeom
        except AttributeError:
            raise AttributeError(
                "Attribute does not exist.  'preoptgeom' should be specified at first")

    def preoptgeom(self, obj=None):
        """
        Args:
            obj (Optgeom): An Optgeom block object
        """
        import warnings

        warnings.warn(
            "Keyword 'PREOPTGEOM' is launched. To set geometric optimization keywords, use 'self.optgeom' attribute.")

        conflict = ['_block_optgeom', '_nooptgeom']
        super(Freqcalc, self).clean_conflict('_block_optgeom', conflict)

        if obj == None:  # New obj
            self._block_optgeom = Optgeom()
            self._block_optgeom._block_bg = 'PREOPTGEOM\n'
            self._block_optgeom._block_ed = 'END\n'
        elif obj == '':
            self._block_optgeom.clean_block()
        else:
            self._block_optgeom = obj
            self._block_optgeom._block_bg = 'PREOPTGEOM\n'
            self._block_optgeom._block_ed = 'END\n'

    def dispersion(self, key='DISPERSION'):
        self._dispersion = super(Freqcalc, self).assign_keyword(key, [])

    def bands(self, ISS=None, NSUB=None, NLINE=None, points=[]):
        if ISS == None:
            self._bands = super(Freqcalc, self).assign_keyword('BANDS', [])
        elif ISS == '':
            self._bands = super(Freqcalc, self).assign_keyword('', [])
        else:
            shape, value = super(Freqcalc, self).set_list(NLINE, points)
            self._bands = super(Freqcalc, self).assign_keyword('BANDS', [2, ] + shape, [ISS, NSUB] + value)

    def modes(self, key='MODES'):
        self._modes = super(Freqcalc, self).assign_keyword(key, [])

    def nomodes(self, key='NOMODES'):
        self._modes = super(Freqcalc, self).assign_keyword(key, [])

    def numderiv(self, N=None):
        self._numderiv = super(Freqcalc, self).assign_keyword(
            'NUMDERIV', [1, ], N)

    def pressure(self, NP=None, P1=None, P2=None):
        self._pressure = super(Freqcalc, self).assign_keyword(
            'PRESSURE', [3, ], [NP, P1, P2])

    def restart(self, key='RESTART'):
        self._restart = super(Freqcalc, self).assign_keyword('RESTART', [])

    def stepsize(self, STEP=None):
        self._stepsize = super(Freqcalc, self).assign_keyword('NUMDERIV', [1, ], STEP)

    def temperat(self, NT=None, T1=None, T2=None):
        self._temperat = super(Freqcalc, self).assign_keyword('TEMPERAT', [3, ], [NT, T1, T2])


class BasisSet(BlockBASE):
    """
    Basis Set block object
    """

    def __init__(self):
        self._block_bg = ''
        self._block_ed = 'ENDBS\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'BASISSET' : '_basisset',
            'GHOSTS'   : '_ghosts',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    def basisset(self, NAME=None):
        self._basisset = super(BasisSet, self).assign_keyword(
            'BASISSET', [1, ], NAME)
        if NAME == '':
            self._block_ed = 'ENDBS\n'
        else: # Otherwise _block_bg and _block_ed = '', this block would be recoginzed as an empty block
            self._block_ed = None

    def from_bse(self, name, element, filename=None, append=False):
        """
        Download basis set definitions from `Basis Set Exchange <https://www.basissetexchange.org/>`_.

        Args:
            name (str): Basis set's name.
            element (list[str] | list[int] | list[list[str | int, int]]): List
                of elements, specified by either atomic number or label. When
                a nelement\*2 list is used, the second entry is recognized as
                conventional atomic numbers.
            filename (None | str): If not None, print basis set definitions to
                a text file
            append (bool): Whether to cover old entries. If the old entry
                contains 'BASISSET', it will be removed anyway.
        """
        from CRYSTALClear.base.basisset import BasisSetBASE

        self._check_bs(append)
        if type(element[0]) == list or type(element[0]) == tuple:
            element_real = [i[0] for i in element]
            zconv = [i[1] for i in element]
            bs_obj = BasisSetBASE.from_bse(name, element_real, zconv)
        else:
            bs_obj = BasisSetBASE.from_bse(name, element)

        self._basisset += bs_obj.data

        if filename != None:
            bs_obj.to_file(file=filename)

    def from_string(self, string, fmt='crystal', filename=None, append=False):
        """
        Basis set from a string

        Args:
            string (str): A line of string. Use '\\n' to break lines. The ending
                line '99 0' is needed but not 'END'.
            fmt (str): Format of basis set string. if not 'crystal', this
                method calls `Basis Set Exchange API <https://molssi-bse.github.io/basis_set_exchange/index.html>`_ to convert it.
            filename (None | str): See ``from_bse``
            append (bool): See ``from_bse``
        """
        from CRYSTALClear.base.basisset import BasisSetBASE

        self._check_bs(append)
        bs_obj = BasisSetBASE.from_string(string, fmt)
        self._basisset += bs_obj.data

        if filename != None:
            bs_obj.to_file(file=filename)

    def from_file(self, file, fmt='crystal', filename=None, append=False):
        """
        Basis set from a file

        Args:
            file (file): A formatted text file with basis set definitions. The
                ending line '99 0' is needed but not 'END'.
            fmt (str): Format of basis set string. if not 'crystal', this
                method calls `Basis Set Exchange API <https://molssi-bse.github.io/basis_set_exchange/index.html>`_ to convert it.
            filename (None | str): See ``from_bse``
            append (bool): See ``from_bse``
        """
        from CRYSTALClear.base.basisset import BasisSetBASE

        self._check_bs(append)
        bs_obj = BasisSetBASE.from_file(file, fmt)
        self._basisset += bs_obj.data

        if filename != None:
            bs_obj.to_file(file=filename)

    def from_obj(self, bs_obj, filename=None, append=False):
        """
        Define basis set from a BasisSetBASE object.

        Args:
            bs_obj (BasisSetBASE): A CRYSTALClear.base.basisset.BasisSetBASE
                object.
            filename (None | str): See ``from_bse``
            append (bool): See ``from_bse``
        """
        self._check_bs(append)
        self._basisset += bs_obj.data

        if filename != None:
            bs_obj.to_file(file=filename)

    def _check_bs(self, append):
        """
        Check basis set definitions. Not an independent method that designed to
        be called.
        """
        import warnings
        import re

        if hasattr(self, '_basisset') and append == False:
            warnings.warn('The previous basis set will be erased.', stacklevel=2)
            self._basisset = ''
            if 'BASISSET' in self._basisset:
                self._block_ed = 'ENDBS\n'
        elif hasattr(self, '_basisset') and append == True:
            if 'BASISSET' in self._basisset:
                warnings.warn("'BASISSET' detected. The previous definition will be erased anyway.",
                              stacklevel=2)
                self._basisset = ''
                self._block_ed = 'ENDBS\n'
            else:
                self._basisset = re.sub(r'99\s+0\n$', "", self._basisset)
        else:
            self._basisset = ''

    def ghosts(self, NA=None, LA=[]):
        shape, value = super(BasisSet, self).set_list(NA, LA)
        self._ghosts = super(BasisSet, self).assign_keyword(
            'GHOSTS', shape, value)


class SCF(BlockBASE):
    """
    SCF block object
    """

    def __init__(self):
        self._block_bg = ''
        self._block_ed = 'ENDSCF\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'FIXINDEX' : '_fixbg',
            'DFT'      : 'dft',  # DFT sub-block
            'DFTD3'    : 'dftd3',  # DFTD3 sub-block
            'GCP'      : 'gcp',  # GCP sub-block
            'GCPAUTO'  : '_gcpauto',
            'SMEAR'    : '_smear',
            'ATOMSPIN' : '_atomspin',
            'TOLDEE'   : '_toldee',
            'DIIS'     : '_diis',
            'NODIIS'   : '_diis',
            'DIISALLK' : '_diisallk',
            'HISTDIIS' : '_histdiis',
            'PRTDIIS'  : '_prtdiis',
            'MAXCYCLE' : '_maxcycle',
            'GUESSP'   : '_guessp',
            'FMIXING'  : '_fmixing',
            'NOBIPOLA' : '_nobipola',
            'NOBIPCOU' : '_nobipcou',
            'NOBIPEXC' : '_nobipexc',
            'TOLINTEG' : '_tolinteg',
            'LDREMO'   : '_ldremo',
            'BIPOSIZE' : '_biposize',
            'EXCHSIZE' : '_exchsize',
            'SHRINK'   : '_shrink',
            'EXCHANGE' : '_exchange',
            'POSTSCF'  : '_postscf',
            'PPAN'     : '_ppan',
            'GRADCAL'  : '_gradcal',
            'CMPLXFAC' : '_cmplxfac',
            'REPLDATA' : '_repldata',
            'STDIAG'   : '_stdiag',
            'GEOM'     : 'fixgeom',  # FIXINDEX - GEOM subblock. Must be put at the end
            'BASE'     : 'fixbase',  # FIXINDEX - BASE subblock. Must be put at the end. GEBA subblock not supported
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    @property
    def dft(self):
        """
        Subblock object DFT
        """
        if not hasattr(self, '_block_dft'):
            self.set_dft()
        return self._block_dft

    def set_dft(self, obj=None):
        """
        DFT subblock

        Args:
            obj (DFT | str): A block object of 'DFT' submodule. Or a string in
                CRYSTAL d12 format
        """
        from CRYSTALClear.base.crysd12 import DFT

        self._block_dft = DFT()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_dft.clean_block()
            else:
                self._block_dft.analyze_text(obj)
        else:
            self._block_dft = obj

    @property
    def dftd3(self):
        """
        Subblock object DFTD3
        """
        if not hasattr(self, '_block_dftd3'):
            self.set_dftd3()
        return self._block_dftd3

    def set_dftd3(self, obj=None):
        """
        DFTD3 subblock

        Args:
            obj (DFTD3 | str): A block object of 'DFTD3' submodule. Or a string
                in CRYSTAL d12 format
        """
        from CRYSTALClear.base.crysd12 import DFTD3

        self._block_dftd3 = DFTD3()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_dftd3.clean_block()
            else:
                self._block_dftd3.analyze_text(obj)
        else:
            self._block_dftd3 = obj

    @property
    def gcp(self):
        """
        Subblock object GCP
        """
        if not hasattr(self, '_block_gcp'):
            self.set_gcp()
        return self._block_gcp

    def set_gcp(self, obj=None):
        """
        GCP subblock

        Args:
            obj (GCP | str): A block object of 'GCP' submodule. Or a string in
                CRYSTAL d12 format.
        """
        from CRYSTALClear.base.crysd12 import GCP

        self._block_gcp = GCP()
        if obj == None:  # Initialize block
            return
        elif type(obj) == str:
            if obj == '': # Clean data
                self._block_gcp.clean_block()
            else:
                self._block_gcp.analyze_text(obj)
        else:
            self._block_gcp = obj

    @property
    def fixgeom(self):
        try:
            return self._block_fixgeom
        except AttributeError:
            raise AttributeError(
                "Attribute does not exist. 'fixindex' is not defined.")

    @property
    def fixbase(self):
        try:
            return self._block_fixbase
        except AttributeError:
            raise AttributeError(
                "Attribute does not exist. 'fixindex' is not defined.")

    def fixindex(self, key1=None, obj1=None, obj2=None):
        """
        Args:
            key1 (str): 'GEOM', 'BASE' or 'GEBA'. Fixindex block keywords
            obj1 (Geom | BasisSet): Geometry or basis set object.
            obj2 (BasisSet): *key1 = GEBA only*. Basis set object.
        """
        import warnings

        if key1 == None:
            self._fixbg = super(SCF, self).assign_keyword('FIXINDEX', [])

        elif key1 == '':
            self._fixbg = super(SCF, self).assign_keyword('FIXINDEX', [], '')
            self._block_ed = 'ENDSCF\n'
            if hasattr(self, '_block_fixgeom'):
                self._block_fixgeom.clean_block()
            elif hasattr(self, '_block_fixbase'):
                self._block_fixbase.clean_block()

        elif key1 == 'GEOM':
            warnings.warn( "'GEOM' keyword of 'FIXINDEX' is identified. Use 'fixgeom' for attributes of the geometry subblock.",
                          stacklevel=2)
            self._fixbg = super(SCF, self).assign_keyword('FIXINDEX', [])
            self._block_fixgeom = Geom()
            if obj1 == None:
                self._block_fixgeom._block_bg = 'ENDSCF\nGEOM\n'
                self._block_fixgeom._block_ed = 'END\n'
                self._block_ed = None # Not '', or the block would be recoginized as an empty one
            elif obj1 == '':
                self._block_fixgeom.clean_block()
                self._block_ed = 'ENDSCF\n'
            else:
                self._block_fixgeom = obj1
                self._block_fixgeom._block_bg = 'ENDSCF\nGEOM\n'
                self._block_fixgeom._block_ed = 'END\n'
                self._block_ed = None # Not '', or the block would be recoginized as an empty one

        elif key1 == 'BASE':
            warnings.warn("'BASE' keyword of 'FIXINDEX' is identified. Use 'fixbase' for attributes of the basis set subblock.",
                          stacklevel=2)
            self._fixbg = super(SCF, self).assign_keyword('FIXINDEX', [])
            self._block_fixbase = BasisSet()
            if obj1 == None:
                self._block_fixbase._block_bg = 'ENDSCF\nBASE\n'
                self._block_fixbase._block_ed = 'END\n'
                self._block_ed = None # Not '', or the block would be recoginized as an empty one
            elif obj1 == '':
                self._block_fixbase.clean_block()
                self._block_ed = 'ENDSCF\n'
            else:
                self._block_fixbase = obj1
                self._block_fixbase._block_bg = 'ENDSCF\nBASE\n'
                self._block_fixbase._block_ed = 'END\n'
                self._block_ed = None # Not '', or the block would be recoginized as an empty one

        # GEBA subblock not supported
        # elif key1 == 'GEBA':
        #     warnings.warn("'GEBA' keyword of 'FIXINDEX' is identified. Use 'fixgeom' for attributes of the geometry subblock and 'fixbase' for attributes of the basis set subblock.")
        #     self._block_fixgeom = Geom()
        #     self._block_fixbase = BasisSet()
        #     if obj1 == None:
        #         self._block_fixgeom._block_bg = 'GEBA\n'
        #         self._block_fixgeom._block_ed = ''
        #         self._block_fixbase._block_bg = ''
        #         self._block_fixbase._block_ed = 'END\n'
        #     elif obj1 == '':
        #         self._block_fixgeom.clean_block()
        #         self._block_fixbase.clean_block()
        #     else:
        #         self._block_fixgeom = obj1
        #         self._block_fixgeom._block_bg = 'GEBA\n'
        #         self._block_fixgeom._block_ed = ''
        #         self._block_fixbase = obj2
        #         self._block_fixbase._block_bg = ''
        #         self._block_fixbase._block_ed = 'END\n'

        else:
            raise ValueError('Keyword error. Allowed keywords: GEOM, BASE. GEBA not supported.')

    def biposize(self, ISIZE=None):
        self._biposize = super(SCF, self).assign_keyword('BIPOSIZE', [1, ], ISIZE)

    def exchsize(self, ISIZE=None):
        self._exchsize = super(SCF, self).assign_keyword('EXCHSIZE', [1, ], ISIZE)

    def toldee(self, ITOL=None):
        self._toldee = super(SCF, self).assign_keyword('TOLDEE', [1, ], ITOL)

    def guessp(self, key='GUESSP'):
        self._guessp = super(SCF, self).assign_keyword(key, [])

    def atomspin(self, NA=None, LA=[]):
        shape, value = super(SCF, self).set_list(NA, LA)
        self._atomspin = super(SCF, self).assign_keyword('ATOMSPIN', shape, value)

    def tolinteg(self, ITOL1=None, ITOL2=None, ITOL3=None, ITOL4=None, ITOL5=None):
        conflict = ['_tolinteg', '_nobipola', '_nobipcou', '_nobipexc']
        super(SCF, self).clean_conflict('_tolinteg', conflict)
        self._tolinteg = super(SCF, self).assign_keyword(
            'TOLINTEG', [5, ], [ITOL1, ITOL2, ITOL3, ITOL4, ITOL5]
        )

    def nobipola(self, key='NOBIPOLA'):
        conflict = ['_tolinteg', '_nobipola', '_nobipcou', '_nobipexc']
        super(SCF, self).clean_conflict('_nobipola', conflict)
        self._nobipola = super(SCF, self).assign_keyword(key, [])

    def nobipcou(self, key='NOBIPCOU'):
        conflict = ['_tolinteg', '_nobipola', '_nobipcou', '_nobipexc']
        super(SCF, self).clean_conflict('_nobipcou', conflict)
        self._nobipcou = super(SCF, self).assign_keyword(key, [])

    def nobipexc(self, key='NOBIPEXC'):
        conflict = ['_tolinteg', '_nobipola', '_nobipcou', '_nobipexc']
        super(SCF, self).clean_conflict('_nobipexc', conflict)
        self._nobipexc = super(SCF, self).assign_keyword(key, [])

    def ldremo(self, value):
        self._ldremo = super(SCF, self).assign_keyword('LDREMO', [1, ], value)

    def maxcycle(self, MAX=None):
        self._maxcycle = super(SCF, self).assign_keyword('MAXCYCLE', [1, ], MAX)

    def fmixing(self, IPMIX=None):
        self._maxcycle = super(SCF, self).assign_keyword('FMIXING', [1, ], IPMIX)

    def shrink(self, IS=None, ISP=None, IS1=None, IS2=None, IS3=None):
        if IS1 == None:
            self._shrink = super(SCF, self).assign_keyword('SHRINK', [2, ], [IS, ISP])
        else:
            self._shrink = super(SCF, self).assign_keyword('SHRINK', [2, 3],
                                                           [IS, ISP, IS1, IS2, IS3])

    def gcpauto(self, key='GCPAUTO'):
        self._gcpauto = super(SCF, self).assign_keyword(key, [])

    def smear(self, WIDTH=None):
        self._smear = super(SCF, self).assign_keyword('SMEAR', [1, ], WIDTH)

    def ppan(self, key='PPAN'):
        self._ppan = super(SCF, self).assign_keyword(key, [])

    def gradcal(self, key='GRADCAL'):
        self._gradcal = super(SCF, self).assign_keyword(key, [])

    def exchange(self, key='EXCHANGE'):
        self._exchange = super(SCF, self).assign_keyword(key, [])

    def postscf(self, key='POSTSCF'):
        self._postscf = super(SCF, self).assign_keyword(key, [])

    def diis(self, key='DIIS'):
        self._diis = super(SCF, self).assign_keyword(key, [])

    def nodiis(self, key='NODIIS'):
        self._diis = super(SCF, self).assign_keyword(key, [])

    def diisallk(self, key='DIISALLK'):
        import warnings

        if hasattr(self, '_diis'):
            if 'NODIIS' in self._diis:
                warnings.warn("Keyword 'NODIIS' is set. It will be removed.",
                              stacklevel=2)
                self._diis = ''
        self._diisallk = super(SCF, self).assign_keyword(key, [])

    def histdiis(self, NCYC=None):
        import warnings

        if hasattr(self, '_diis'):
            if 'NODIIS' in self._diis:
                warnings.warn("Keyword 'NODIIS' is set. It will be removed.",
                              stacklevel=2)
                self._diis = ''
        self._histdiis = super(SCF, self).assign_keyword('HISTDIIS', [1,], NCYC)

    def prtdiis(self, key='PRTDIIS'):
        import warnings

        if hasattr(self, '_diis'):
            if 'NODIIS' in self._diis:
                warnings.warn("Keyword 'NODIIS' is set. It will be removed.",
                              stacklevel=2)
                self._diis = ''
        self._prtdiis = super(SCF, self).assign_keyword(key, [])

    def cmplxfac(self, WEIGHT=None):
        self._cmplxfac = super(SCF, self).assign_keyword('CMPLXFAC', [1, ], WEIGHT)

    def repldata(self, key='REPLDATA'):
        self._repldata = super(SCF, self).assign_keyword(key, [])

    def stdiag(self, key='STDIAG'):
        self._stdiag = super(SCF, self).assign_keyword(key, [])


class DFT(BlockBASE):
    """
    DFT block object
    """

    def __init__(self):
        self._block_bg = 'DFT\n'
        self._block_ed = 'ENDDFT\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'SPIN'     : '_spin',
            'EXCHANGE' : '_exchange',
            'CORRELAT' : '_correlat',
            'OLDGRID'  : '_gridsz',
            'LGRID'    : '_gridsz',
            'XLGRID'   : '_gridsz',
            'XXLGRID'  : '_gridsz',
            'XXXLGRID' : '_gridsz',
            'RADIAL'   : '_gridr',
            'ANGULAR'  : '_grida',
            'SVWN'     : '_xcfunc', # Standalone keywords for text analysis - I know it's a stupid idea - to ensure interactive i/o
            'BLYP'     : '_xcfunc',
            'PBEXC'    : '_xcfunc',
            'PBESOLXC' : '_xcfunc',
            'SOGGAXC'  : '_xcfunc',
            'SOGGA11'  : '_xcfunc',
            'B3PW'     : '_xcfunc',
            'B3LYP'    : '_xcfunc',
            'PBE0'     : '_xcfunc',
            'PBESOL0'  : '_xcfunc',
            'B1WC'     : '_xcfunc',
            'WC1LYP'   : '_xcfunc',
            'B97H'     : '_xcfunc',
            'PBE0-13'  : '_xcfunc',
            'SOGGA11X' : '_xcfunc',
            'MPW1PW91' : '_xcfunc',
            'MPW1K'    : '_xcfunc',
            'HSE06'    : '_xcfunc',
            'HSESOL'   : '_xcfunc',
            'SC-BLYP'  : '_xcfunc',
            'HISS'     : '_xcfunc',
            'RSHXLDA'  : '_xcfunc',
            'WB97'     : '_xcfunc',
            'WB97X'    : '_xcfunc',
            'LC-WPBE'  : '_xcfunc',
            'LC-WPBESOL': '_xcfunc',
            'LC-WBLYP' : '_xcfunc',
            'LC-BLYP'  : '_xcfunc',
            'CAM-B3LYP': '_xcfunc',
            'LC-PBE'   : '_xcfunc',
            'M06L'     : '_xcfunc',
            'REVM06L'  : '_xcfunc',
            'MN15L'    : '_xcfunc',
            'SCAN'     : '_xcfunc',
            'R2SCAN'   : '_xcfunc',
            'B1B95'    : '_xcfunc',
            'MPW1B95'  : '_xcfunc',
            'MPW1B1K'  : '_xcfunc',
            'PW6B95'   : '_xcfunc',
            'PWB6K'    : '_xcfunc',
            'M05'      : '_xcfunc',
            'M052X'    : '_xcfunc',
            'M06'      : '_xcfunc',
            'M062X'    : '_xcfunc',
            'M06HF'    : '_xcfunc',
            'MN15'     : '_xcfunc',
            'REVM06'   : '_xcfunc',
            'SCAN0'    : '_xcfunc',
            'R2SCANH'  : '_xcfunc',
            'R2SCAN0'  : '_xcfunc',
            'R2SCAN50' : '_xcfunc',
            'MN15'     : '_xcfunc',
            'BLYP-D3'  : '_xcfunc',
            'PBE-D3'   : '_xcfunc',
            'B97-D3'   : '_xcfunc',
            'B3LYP-D3' : '_xcfunc',
            'PBE0-D3'  : '_xcfunc',
            'PW1PW-D3' : '_xcfunc',
            'M06-D3'   : '_xcfunc',
            'HSE06-D3' : '_xcfunc',
            'HSESOL-D3': '_xcfunc',
            'LC-WPBE-D3': '_xcfunc',
            'B3LYP-D3' : '_xcfunc',
            'PBEH3C'   : '_xcfunc',
            'HSE3C'    : '_xcfunc',
            'B973C'    : '_xcfunc',
            'PBESOL03C': '_xcfunc',
            'HSESOL3C' : '_xcfunc',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    def spin(self, key='SPIN'):
        self._spin = super(DFT, self).assign_keyword(key, [])

    def exchange(self, ex=None):
        if hasattr(self, '_xcfunc'):
            raise AttributeError(
                'Exchange-correlation functional is already set.')
        self._exchange = super(DFT, self).assign_keyword('EXCHANGE', [1, ], ex)

    def correlat(self, cor=None):
        if hasattr(self, '_xcfunc'):
            raise AttributeError('Exchange-correlation functional is already set.')
        self._correlat = super(DFT, self).assign_keyword('CORRELAT', [1, ], cor)

    def xcfunc(self, xc=None):
        if hasattr(self, '_exchange') or hasattr(self, '_correlat'):
            raise AttributeError('Separate keywords are set for exchange / correlation functionals.')
        self._xcfunc = super(DFT, self).assign_keyword(None, [1, ], xc)

    def lgrid(self, key='LGRID'):
        self._gridsz = super(DFT, self).assign_keyword(key, [])

    def oldgrid(self, key='OLDGRID'):
        self._gridsz = super(DFT, self).assign_keyword(key, [])

    def xlgrid(self, key='XLGRID'):
        self._gridsz = super(DFT, self).assign_keyword(key, [])

    def xxlgrid(self, key='XXLGRID'):
        self._gridsz = super(DFT, self).assign_keyword(key, [])

    def xxxlgrid(self, key='XXXLGRID'):
        self._gridsz = super(DFT, self).assign_keyword(key, [])

    def radial(self, NR=None, RL=[], IL=[]):
        if hasattr(self, '_gridsz'):
            raise AttributeError("Pre-defined integrated grid '{}' is defined.".format(self._gridsz[:-1]))
        if NR != None and NR != '':
            if len(RL) != len(IL) and NR != len(RL):
                raise ValueError('Inconsistent definition of parameters.')
        self._gridr = super(DFT, self).assign_keyword('RADIAL', [1, len(RL), len(IL)], [NR, ] + RL + IL)

    def angular(self, NI=None, AL=[], LEV=[]):
        if hasattr(self, '_gridsz'):
            raise AttributeError("Pre-defined integrated grid '{}' is defined.".format(self._gridsz[:-1]))
        if NI != None and NI != '':
            if len(AL) != len(LEV) and NI != len(AL):
                raise ValueError('Inconsistent definition of parameters.')
        self._grida = super(DFT, self).assign_keyword('ANGULAR', [1, len(AL), len(LEV)], [NI, ] + AL + LEV)


class DFTD3(BlockBASE):
    """
    DFTD3 block object
    """

    def __init__(self):
        self._block_bg = 'DFTD3\n'
        self._block_ed = 'END\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'VERSION'   : '_version',
            'FUNC'      : '_func',
            'ABC'       : '_abc',
            'S6'        : '_s6',
            'S8'        : '_s8',
            'A1'        : '_a1',
            'A2'        : '_a2',
            'RS6'       : '_rs6',
            'RS8'       : '_rs8',
            'RADIUS'    : '_radius',
            'CNRADIUS'  : '_cnradius',
            'ABCRADIUS' : '_abcradius',
            'PRINTC6'   : '_printc6',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    def version(self, NAT=None):
        self._version = super(DFTD3, self).assign_keyword('VERSION', [1, ], NAT)

    def func(self, CHAR=None):
        self._func = super(DFTD3, self).assign_keyword('FUNC', [1, ], CHAR)

    def abc(self, key='ABC'):
        self._abc = super(DFTD3, self).assign_keyword(key, [])

    def s6(self, s6=None):
        self._s6 = super(DFTD3, self).assign_keyword('S6', [1, ], s6)

    def s8(self, s8=None):
        self._s8 = super(DFTD3, self).assign_keyword('S8', [1, ], s8)

    def a1(self, a1=None):
        self._a1 = super(DFTD3, self).assign_keyword('A1', [1, ], a1)

    def a2(self, a1=None):
        self._a2 = super(DFTD3, self).assign_keyword('A2', [1, ], a2)

    def rs6(self, rs6=None):
        self._rs6 = super(DFTD3, self).assign_keyword('RS6', [1, ], rs6)

    def rs8(self, rs8=None):
        self._rs8 = super(DFTD3, self).assign_keyword('RS8', [1, ], rs8)

    def radius(self, radius=None):
        self._radius = super(DFTD3, self).assign_keyword('RADIUS', [1, ], radius)

    def cnradius(self, cnradius=None):
        self._cnradius = super(DFTD3, self).assign_keyword('CNRADIUS', [1, ], cnradius)

    def abcradius(self, abcradius=None):
        self._abcradius = super(DFTD3, self).assign_keyword('ABCRADIUS', [1, ], abcradius)

    def printc6(self, key='PRINTC6'):
        self._printc6 = super(DFTD3, self).assign_keyword(key, [])


class GCP(BlockBASE):
    """
    GCP block object
    """

    def __init__(self):
        self._block_bg = 'GCP\n'
        self._block_ed = 'END\n'
        self._block_data = ''
        self._block_dict = {  # The sequence of keywords should follow rules in the manual
            'METHOD'     : '_method',
            'SIGMA'      : '_sigma',
            'ALPHA'      : '_alpha',
            'BETA'       : '_beta',
            'ETA'        : '_eta',
            'RADIUS'     : '_radius',
            'PRINTEMISS' : '_printemiss',
        }
        key = list(self._block_dict.keys())
        attr = list(self._block_dict.values())
        self._block_key = sorted(set(key), key=key.index)
        self._block_attr = sorted(set(attr), key=attr.index)

    def method(self, method=None):
        self._method = super(GCP, self).assign_keyword('METHOD', [1, ], method)

    def sigma(self, sigma=None):
        self._sigma = super(GCP, self).assign_keyword('SIGMA', [1, ], sigma)

    def alpha(self, alpha=None):
        self._alpha = super(GCP, self).assign_keyword('ALPHA', [1, ], alpha)

    def beta(self, beta=None):
        self._beta = super(GCP, self).assign_keyword('BETA', [1, ], beta)

    def eta(self, eta=None):
        self._eta = super(GCP, self).assign_keyword('ETA', [1, ], eta)

    def radius(self, radius=None):
        self._radius = super(GCP, self).assign_keyword('RADIUS', [1, ], radius)

    def printemiss(self, key='PRINTEMISS'):
        self._printemiss = super(GCP, self).assign_keyword(key, [])



from CRYSTALClear.base import basisset
from CRYSTALClear.base import inputbase
from CRYSTALClear.base import crysd12
from CRYSTALClear.base import crysout
from CRYSTALClear.base import propout
from CRYSTALClear.base import plotbase



class BS_GTF():
    """
    A Gaussian type function object.

    Args:
        info (str): GTF radial coefficients in CRYSTAL format
    """

    def __init__(self, info):
        info = info.replace('D', 'E').split()
        if len(info) == 2:
            self.exp = eval(info[0])
            self.contr = eval(info[1])
            self.pcoef = None
        elif len(info) == 3:
            self.exp = eval(info[0])
            self.contr = eval(info[1])
            self.pcoef = eval(info[2])
        else:
            raise ValueError('Input format error.')

    @property
    def data(self):
        """
        Print basis set information in CRYSTAL format
        """
        return self._print_data()

    def _print_data(self):
        """
        Return to a string in CRYSTAL format
        """
        if self.pcoef == None:
            bs_str = '  {: 16.10E}    {: 16.10E}\n'.format(
                self.exp, self.contr)
        else:
            bs_str = '  {: 16.10E}    {: 16.10E}    {: 16.10E}\n'.format(
                self.exp, self.contr, self.pcoef)

        return bs_str


class BS_Shell():
    """
    The Shell object to set a general basis set.

    Args:
        info (list[str]): String in Crystal format

    Other arguments are consistent with CRYSTAL manual.
    """

    def __init__(self, ITYB, LAT, NG, CHE, SCAL, info):

        from CRYSTALClear.base.basisset import BS_GTF

        if ITYB != 0:
            raise ValueError(
                'BS_Shell class is limited to general definitions of basis set.')

        self.type = ITYB
        self.angular = LAT
        self.ngtf = NG
        self.chg = CHE
        self.scale = SCAL
        self.gtf = []
        for line in info:
            self.gtf.append(BS_GTF(line))

    @property
    def data(self):
        """
        Print basis set information in CRYSTAL format
        """
        return self._print_data()

    def _print_data(self):
        """
        Return to a string in CRYSTAL format
        """
        bs_str = ''
        bs_str += format('%-2i%-2i%-3i%-6.2f%-6.2f\n' %
                         (self.type, self.angular, self.ngtf, self.chg, self.scale))
        for g in self.gtf:
            bs_str += g.data

        return bs_str


class BS_Atom():
    """
    The atom object to set a basis set.

    .. note::

        The type of basis set should be consistent for all the shells.
        Otherwise unexpected errors might happen.

        Free effective core pseudopotential (ECP) definition not supported.

    Args:
        z (int): Convential atomic number Z.
        info (list(str)): String in Crystal BS format
    """

    def __init__(self, z, info):
        import re
        from CRYSTALClear.base.basisset import BS_Shell

        if z > 99:
            self.z = z % 100
        else:
            self.z = z
        self.conventional_atomic_number(z)
        self.nshell = 0
        self.shell = []
        self.ecp = None
        set_stdbs = False
        info.append('')  # Avoid index out of range error
        for idx, line in enumerate(info):
            if re.match(r'^[0-2]\s+[0-5]\s+[0-9]+\s+[0-9,\.,\-,\+]+\s+[0-9,\.,\-,\+]+$', line):
                self.nshell += 1
                data = line.split()
                ITYB = int(data[0])
                LAT = int(data[1])
                NG = int(data[2])
                CHE = float(data[3])
                SCAL = float(data[4])
                if ITYB == 0:  # General definition of basis sets
                    self.shell.append(
                        BS_Shell(ITYB, LAT, NG, CHE, SCAL,
                                 info[idx + 1: idx + NG + 1])
                    )
                else:  # Standard STO-3G / 3(6)-21G* basis sets
                    set_stdbs = True
                    break
            elif re.match(r'^[A-Z]+', line): # ECP line
                if re.match(r'^INPUT', line):
                    raise ValueError('Free-format ECP not supported')
                self.ecp = line
            else:
                continue

        if set_stdbs == True:
            self._set_stdbs(info)

    def conventional_atomic_number(self, zconv):
        """
        Set convential atomic number.

        Args:
            zconv (int): Z(real atomic number) + n\*100. N is an integer.
        """
        if zconv % 100 != self.z:
            raise ValueError(
                'Convential atomic number should be Z + n*100. N is an integer.')

        self.zconv = zconv
        return

    def _set_stdbs(self, z, info):
        """
        Set the standard STO-3G / 3(6)-21G* basis sets. All the settings will
        be converted to general BS definitions.

        Args:
            info (list(str)): String in Crystal BS format
        """
        import warnings
        import copy
        try:
            import basis_set_exchange as bse
            print("module bse imported")
        except ImportError:
            import sys
            print (sys.exc_info())
            print("module basis_set_exchange unavailable: please install to activate")
#        import basis_set_exchange as bse

        from CRYSTALClear.base.basisset import Basisset

        warnings.warn('''The built-in standard STO-3G / 3(6)-21G basis set is used.
CRYSTALClear will substract basis set definition from Basis Set Exchange.
That might lead to discripencies in basis set definitions.''')

        # set a general 1-atom BS template
        ITYB_init = info[1].split()[0]
        NG_init = info[1].split()[2]
        if ITYB_init == 1:  # STO-3G
            bs = Basisset('STO-3G', [self.z, ])
        elif ITYB_init == 2 and NG_init == 3:  # 3-21G
            bs = Basisset('3-21G', [self.z, ])
        elif ITYB_init == 2 and NG_init == 6:  # 6-21G
            bs = Basisset('6-21G', [self.z, ])
        else:
            raise ValueError('Basis set input format error.')

        atom = bs.atom[self.z]
        self.nshell = copy.deepcopy(atom.nshell)
        self.shell = copy.deepcopy(atom.shell)
        del atom, bs
        # Update object with user defined value. Every line defines a shell
        for idx, line in enumerate(info):
            data = line.split()
            if len(data) == 2:
                continue
            if self.shell[idx].angular != int(data[1]) or self.shell[idx].ngtf != int(data[2]):
                warnings.warn(
                    'The sequence of input BS is not consistent with BSE data. Trying to find the matched shell...')
                found_flag = False
                for j in range(len(info)):
                    data2 = info[j].split()
                    if self.shell[j].angular == int(data2[1]) or self.shell[j].ngtf != int(data2[2]):
                        self.shell[j].chg = float(data2[3])
                        self.shell[j].scale = float(data2[4])
                        found_flag = True
                        break
                if found_flag == False:
                    raise Exception(
                        'The input BS is not a correct STO-3G / 3(6)-21G basis set.')
            else:
                self.shell[idx].chg = float(data[3])
                self.shell[idx].scale = float(data[4])
        return

    @property
    def data(self):
        """
        Print basis set information in CRYSTAL format
        """
        return self._print_data()

    def _print_data(self):
        """
        Return to a string in CRYSTAL format

        Returns:
            bs_str (str): Formatted basis set string of an atom.
        """
        bs_str = ''
        bs_str += format('%-5i%-5i\n' % (self.zconv, self.nshell))
        if self.ecp != None:
            bs_str += format('%s\n' % self.ecp)

        for s in self.shell:
            bs_str += s.data

        return bs_str


class BasisSetBASE():
    """
    The basisset object in CRYSTAL format. When called, one can pass the
    following arguments:

    Args:
        name (str | None): When specified, download the corresponding basis set
            from `Basis Set Exchange(BSE) <https://www.basissetexchange.org/>`_
        element (list[str] | list[int]): Elements to download. Either as a list
            of atomic numbers or labels.

    .. note::

        For basis sets from BSE, reference and roles are omitted.

    """

    def __init__(self):
        pass

    @classmethod
    def from_bse(cls, name, element, zconv=None):
        """
        Download basis set definitions from BSE.

        Args:
            name (str): Basis set's name.
            element (list[str] | list[int]): List of elements.
        """
        import basis_set_exchange as bse

        bs_str = bse.get_basis(name, elements=element, fmt='crystal')

        return cls()._set_atom(bs_str, zconv)

    @classmethod
    def from_string(cls, bs_str, fmt='crystal'):
        """
        Define basis set from a string.

        Args:
            bs_str (str)
            fmt (str): Format string. Consistent with BSE python API.
        """
        import re
        import basis_set_exchange as bse

        if not re.match(r'^crystal$', fmt, re.IGNORECASE):
            bs_str = bse.convert_formatted_basis_str(bs_str, fmt, 'crystal')

        return cls()._set_atom(bs_str)

    @classmethod
    def from_file(cls, file, fmt='crystal'):
        """
        Define a basis set from a file.
        """
        import re
        try:
            import basis_set_exchange as bse
            print("module bse imported")
        except ImportError:
            import sys
            print (sys.exc_info())
            print("module basis_set_exchange unavailable: please install to activate")
#        import basis_set_exchange as bse

        bs_file = open(file, 'r')
        bs_str = bs_file.read()
        bs_file.close()
        if not re.match(r'^crystal$', fmt, re.IGNORECASE):
            bs_str = bse.convert_formatted_basis_str(bs_str, fmt, 'crystal')

        return cls()._set_atom(bs_str)

    def _set_atom(self, info, zconv=None):
        """
        Assign basis set parameters at atom level.

        Args:
            info (str): String in Crystal format
            zconv (list[int]): If not none, use the conventional atomic number.
                Its length must be nelement. Its sequence must be consistent
                with basis set's
        """
        import re
        from CRYSTALClear.base.basisset import BS_Atom

        info = info.strip().split('\n')
        info = [i.strip() for i in info]

        atom_list = []
        atom_range = []
        for idx, line in enumerate(info):
            if re.match(r'^[0-9]+\s+[0-9]+$', line):
                if re.match(r'^99\s+0$', line):
                    atom_range.append(idx)
                    break
                if zconv == None:
                    z = int(line.split()[0])  # Conventional atomic number
                else:
                    z = zconv[len(atom_range) - 1]
                atom_list.append(z)
                atom_range.append(idx)

        self.atom = {}
        for idx, z in enumerate(atom_list):
            atom_obj = BS_Atom(z, info[atom_range[idx]: atom_range[idx + 1]])
            self.atom[z] = atom_obj

        return self

    @property
    def data(self):
        """
        Print basis set information in CRYSTAL format
        """
        bs_str = ''
        for a in list(self.atom.values()):
            bs_str += a.data
        bs_str += '99   0\n'

        return bs_str

    def to_file(self, file='BASISSET.DAT', fmt='crystal'):
        """
        Print formatted data into a text file.

        Args:
            file (str)
            fmt (str): Output format
        """
        import os
        import re
        import warnings
        try:
            import basis_set_exchange as bse
            print("module bse imported")
        except ImportError:
            import sys
            print (sys.exc_info())
            print("module basis_set_exchange unavailable: please install to activate")
    #    import basis_set_exchange as bse

        f = open(file, 'w+')
        if os.path.isfile(file):
            warnings.warn('File exists. New entry will be attached to the end.')

        bs_str = self.data
        if not re.match(r'^crystal$', fmt, re.IGNORECASE):
            bs_str = bse.convert_formatted_basis_str(bs_str, 'crystal', fmt)

        f.write('%s' % bs_str)
        f.close()



from CRYSTALClear import adsorb
from CRYSTALClear import base
from CRYSTALClear import calculate
from CRYSTALClear import convert
from CRYSTALClear import execute
from CRYSTALClear import geometry
from CRYSTALClear import crystal_io
from CRYSTALClear import plot
from CRYSTALClear import thermodynamics
from CRYSTALClear import units
#from CRYSTALClear import utils

import sys

if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    from importlib_metadata import metadata

__author__ = "CRYSTALClear Development Team"
__email__ = "crystalunito@gmail.com"
__version__ = metadata.version('CRYSTALClear')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base functions for plotting 2D and 3D figures
"""


def plot_cry_bands(bands, k_labels, energy_range, title, not_scaled, mode, linestl,
                   linewidth, color, fermi, k_range, labels, figsize, scheme,
                   sharex, sharey, fermiwidth, fermialpha):
    """
    The base function to plot electron  phonon density of states.

    Args:
        bands (Union[List, object]): List of band objects or a single band object.
        k_labels (Union[List[str], None]): List of strings specifying the labels for high symmetry points along the path.
        energy_range (Union[List[Union[int, float]], None]): List of two integers or floats specifying the energy range (min, max).
        title (Union[str, None]): Title of the plot.
        not_scaled (bool): Flag indicating whether to scale the band structure for comparison.
        mode (str): Mode of the plot ('single', 'multi', 'compare').
        linestl (Union[str, List[str]]): Line style or list of line styles for the bands.
        linewidth (Union[int, List[int]]): Line width or list of line widths for the bands.
        color (Union[str, List[str]]): Color or list of colors for the bands.
        fermi (str): Color of the Fermi level.
        fermiwidth (float): Thickness of the Fermi level
        fermialpha (float): Opacity of the Fermi level
        k_range (Union[List[str], None]): List of two strings specifying the range of k points to plot.
        labels (Union[List[str], None]): List of labels for the bands in multi-mode.
        figsize (Union[Tuple[int, int], None]): Figure size.
        scheme (Union[List[int], Tuple[int, int], None]): Subplot scheme in compare-mode (number of rows, number of columns).
        sharex (Union[bool, str]): Flag or 'row' or 'col' specifying sharing of x-axis.
        sharey (Union[bool, str]): Flag or 'row' or 'col' specifying sharing of y-axis.
        fermialpha(float): Opacity of the fermi level 0-1
        fermiwidth(float): Width of the fermi level

    Raises:
        ValueError: If an invalid mode flag is specified or if there are errors in the input parameters.

    Returns:
        None
    """
    import sys
    import warnings

    import matplotlib.pyplot as plt
    import numpy as np

    greek = {'Alpha': '\u0391', 'Beta': '\u0392', 'Gamma': '\u0393', 'Delta': '\u0394', 'Epsilon': '\u0395', 'Zeta': '\u0396', 'Eta': '\u0397',
             'Theta': '\u0398', 'Iota': '\u0399', 'Kappa': '\u039A', 'Lambda': '\u039B', 'Mu': '\u039C', 'Nu': '\u039D', 'Csi': '\u039E',
             'Omicron': '\u039F', 'Pi': '\u03A0', 'Rho': '\u03A1', 'Sigma': '\u03A3', 'Tau': '\u03A4', 'Upsilon': '\u03A5', 'Phi': '\u03A6',
             'Chi': '\u03A7', 'Psi': '\u03A8', 'Omega': '\u03A9', 'Sigma_1': '\u03A3\u2081'}

    # Error check on the mode flag
    modes = ['single', 'multi', 'compare', 'surface']
    if mode not in modes:
        raise ValueError('The selected mode '+mode+' is not among the possible ones: ' +
                         modes[0]+', ' + modes[1] + ', '+modes[2] + ', or '+modes[3])

    # Automatically activates the multi mode if bands is a list
    if (isinstance(bands, list)) and (mode is modes[0]):
        mode = modes[1]

    # Error chenk on k_label
    if k_labels is not None:

        if isinstance(k_labels, list):
            for element in k_labels:
                if not isinstance(element, str):
                    raise ValueError(
                        'k_label must be a list of strings:' + str(element)+' is not a string')

            if isinstance(bands, list):
                ref = bands[0].n_tick

                for i, file in enumerate(bands):
                    if file.n_tick != ref:
                        i = str(i)
                        raise ValueError(
                            'The ' + i + 'element your file list has a different number of High Symmetry Point!')
            else:
                ref = bands.n_tick

            if len(k_labels) < ref:
                diff = str(ref-len(k_labels))
                raise ValueError('You are lacking ' + diff +
                                 ' High Symmetry Points in k_labels!')

            elif len(k_labels) > ref:
                diff = str(len(k_labels)-ref)
                raise ValueError('You have ' + diff +
                                 'High Symmetry Points in excess in k_labels')

        else:
            raise ValueError('k_labels must be a list of strings')

    # Error check on energy range
    if energy_range != None:

        if isinstance(energy_range, list):

            if len(energy_range) > 2:
                raise ValueError(
                    'energy_range must be a list of two int or float (min,max)')

            for element in energy_range:
                if (not isinstance(element, int)) and (not isinstance(element, float)):
                    raise ValueError('energy_range must be a list of two int or float (min,max): ' + str(
                        element)+', is neither a float or an int')

        else:
            raise ValueError(
                'energy_range must be a list of two int or float (min,max)')

    # Error check on k_range
    if k_range is not None:
        if isinstance(k_range, list):
            if len(k_range) > 2:
                raise ValueError('k_range must be a list of two strings')

            for element in k_range:
                if not isinstance(element, str):
                    raise ValueError(
                        'k_label must be a list of two strings:' + str(element)+' is not a string')

        else:
            raise ValueError('k_range must be a list of two strings')

    # Error check on title
    if title != None:
        if not isinstance(title, str):
            raise ValueError('title needs to be a string')

    if (mode == modes[0]) or (mode == modes[1]) or (mode == modes[3]):

        # plotting of a single band object
        if mode == modes[0]:
            fig = plot_single_cry_bands(
                bands, linestl, linewidth, color, figsize, sharex, sharey)
            ymin = fig[0]
            ymax = fig[1]
            xmin = fig[2]
            xmax = fig[3]
            ax = fig[5]
            fig = fig[4]

        # plot of multiple band objects on a single plot
        elif mode == modes[1]:

            # Error check on the band on the 'multi' mode flag
            if not isinstance(bands, list):
                raise ValueError(
                    'When you choose a ' + modes[1]+' plot bands needs to be a list of band objects')
            # Error check on color for the 'multi' mode flag
            if isinstance(color, list):
                if len(color) != len(bands):
                    raise ValueError(
                        'The number of colors is inconsistent with the number of objects you want to plot')
            else:
                color = ['dimgrey', 'blue', 'indigo', 'slateblue',
                         'thistle', 'purple', 'orchid', 'crimson']

            # Warning comparison with band.spin==2
            for m in bands:
                if m.spin == 2:
                    warnings.warn(
                        "The 'multi' plot is not fully implemented at the moment for file with NSPIN = 2")

            fig = plot_multi_cry_bands(bands,  not_scaled, linestl, linewidth,
                                       color, labels, figsize, sharex, sharey)
            ymin = fig[0]
            ymax = fig[1]
            xmin = fig[2]
            xmax = fig[3]
            ax = fig[5]
            fig = fig[4]

        if mode == modes[3]:
            warnings.warn('The surface bands is not ready yet')
            sys.exit(0)

        # HSP line plot
        if isinstance(bands, list):
            hsp = bands[0].tick_position
        else:
            hsp = bands.tick_position

        if k_labels is not None:
            if len(hsp) != len(k_labels):
                if len(hsp) > len(k_labels):
                    raise ValueError(
                        'You have specified a number of label smalle than the number of High Simmetry Point along the path')
                elif len(hsp) < len(k_labels):
                    raise ValueError(
                        'You have more labels than the High Simmetry point along the path')

        hsp[len(hsp)-1] = xmax

        y_band = np.linspace(ymin*1.05, ymax*1.05, 2)

        for j in hsp:
            x_band = np.ones(2)*j
            ax.plot(x_band, y_band, color='black', linewidth=0.5)

        # plot of the fermi level
        x = np.linspace(xmin, xmax, 2)
        y = np.zeros(2)
        ax.plot(x, y, color=fermi, linewidth=fermiwidth, alpha=fermialpha)

        # definition of the plot title
        if title is not None:
            fig.suptitle(title)

        if not isinstance(bands, list):
            if bands.spin == 2:
                ax.legend()
        else:
            if labels is not None:
                ax.legend()

    # compare mode plot
    elif mode == modes[2]:

        # Error check on type(bands)
        if not isinstance(bands, list):
            raise ValueError(
                'When you choose a ' + modes[2]+' plot bands needs to be a list of band objects')
        # Error check on sharex and sharey option
        if (not isinstance(sharex, bool)) or (not isinstance(sharey, bool)):
            accepted_str = ['row', 'col']

            if (isinstance(sharex, str)) or (isinstance(sharey, str)):
                if sharex not in accepted_str:
                    raise ValueError('sharex can only be equal to row or col')
                elif sharey not in accepted_str:
                    raise ValueError('sharey can only be equal to row or col')
            else:
                raise ValueError('sharex and sharey have to be boolean')

        fig = plot_compare_cry_bands(bands, energy_range, not_scaled, linestl, linewidth,
                                     color, fermi, figsize, scheme, sharex, sharey,
                                     fermiwidth, fermialpha)
        ymin = fig[0]
        ymax = fig[1]
        xmin = fig[2]
        xmax = fig[3]
        hsp = fig[4]
        ax = fig[6]
        fig = fig[5]

    hsp_label = []
    high_sym_point = []
    if k_labels is not None:
        for n in k_labels:
            if n in greek:
                g = greek.get(n)
                hsp_label.append(g)
                high_sym_point.append(n)
            else:
                hsp_label.append(n)
                high_sym_point.append(n)

    # give the possibility through the k_range to select a shorter path than the one calculated
    high_sym_point2 = high_sym_point
    count = 0
    for i in high_sym_point2:
        repeat = high_sym_point2.count(i)
        if repeat != 1:
            for p in range(0, len(high_sym_point2)):
                if p != count:
                    repeat_count = 1
                    q = high_sym_point2[p]
                    r = high_sym_point[p]
                    if (q == i) & (q == r):
                        high_sym_point2[p] = i+str(repeat_count)
                        repeat_count += 1
                        if repeat_count > repeat:
                            repeat_count = 0
        count += 1

    path_dict = dict(zip(high_sym_point2, hsp))

    # definition of the ylim
    if (energy_range is not None) and (sharey == True):
        ymin = energy_range[0]
        ymax = energy_range[1]

    # definition of the xlim
    if k_range is not None:
        xmin = path_dict[k_range[0]]
        xmax = path_dict[k_range[1]]

    if k_labels is not None:
        plt.xticks(hsp, hsp_label)
    else:
        plt.xticks(hsp)

    plt.ylim(ymin, ymax)
    if (mode == modes[0]) or (mode == modes[1]):
        plt.xlim(xmin, xmax)
    elif (mode == modes[2]) and (k_range is not None):
        warnings.warn('The k_range is not available yet for the compare mode')

    return fig, ax


def plot_single_cry_bands(bands, linestl, linewidth, color, figsize, sharex, sharey):

    import matplotlib.pyplot as plt
    import numpy as np

    dx = bands.k_point_plot

    pltband = bands.bands
    no_bands = np.shape(pltband)[0]
    ymin = np.amin(pltband)
    ymax = np.amax(pltband)
    xmin = min(dx)
    xmax = max(dx)
    count1 = 0
    count2 = 0

    # band plot
    if figsize is not None:
        fig, ax = plt.subplots(
            nrows=1, ncols=1, sharex=sharex, sharey=sharey, figsize=figsize)
    else:
        fig, ax = plt.subplots(
            nrows=1, ncols=1, sharex=sharex, sharey=sharey)

    for i in range(no_bands):

        if bands.spin == 1:
            ax.plot(dx, pltband[i, :], color=color,
                    linestyle=linestl, linewidth=linewidth)

        elif bands.spin == 2:
            if count1 == count2:
                ax.plot(dx, pltband[i, :, 0], color='red',
                        linestyle=linestl, linewidth=linewidth, label='Alpha')
                ax.plot(dx, pltband[i, :, 1], color='black',
                        linestyle=linestl, linewidth=linewidth, label='Beta')
            else:
                ax.plot(dx, pltband[i, :, 0], color='red',
                        linestyle=linestl, linewidth=linewidth)
                ax.plot(dx, pltband[i, :, 1], color='black',
                        linestyle=linestl, linewidth=linewidth)
            count1 += 1

    return ymin, ymax, xmin, xmax, fig, ax


def plot_multi_cry_bands(bands,  not_scaled, linestl, linewidth, color,
                         labels, figsize, sharex, sharey):

    import matplotlib.pyplot as plt
    import numpy as np

    # scaling that enables the comparison of band structure calculated at different pressures
    if not_scaled is False:
        reference = xmax = np.amax(bands[0].k_point_plot)
        xmin = np.amin(bands[0].k_point_plot)

    else:
        xmax = []
        xmin = []

    ymin = []
    ymax = []

    if figsize is not None:
        fig, ax = plt.subplots(
            nrows=1, ncols=1, sharex=sharex, sharey=sharey, figsize=figsize)
    else:
        fig, ax = plt.subplots(
            nrows=1, ncols=1, sharex=sharex, sharey=sharey)

    # plot of all the bands obj present in the list
    for index, data in enumerate(bands):
        # scaling that enables the comparison of band structure calculated at different pressures
        if not_scaled is False:
            k_max = np.amax(data.k_point_plot)
            dx = (data.k_point_plot/k_max)*reference
        else:
            dx = data.k_point_plot
            xmin.append(np.amin(dx))
            xmax.append(np.amax(dx))

        pltband = data.bands
        no_bands = np.shape(pltband)[0]
        ymin.append(np.amin(pltband))
        ymax.append(np.amax(pltband))

        count1 = 0
        count2 = 0

        # Effective plot
        for j in range(no_bands):

            if (count1 == count2) and (labels is not None):
                if data.spin == 1:
                    if isinstance(linestl, list):
                        if isinstance(linewidth, list):
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl[index], linewidth=linewidth[index], label=labels[index])
                        else:
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl[index], linewidth=linewidth, label=labels[index])
                    else:
                        if isinstance(linewidth, list):
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl, linewidth=linewidth[index], label=labels[index])
                        else:
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl, linewidth=linewidth, label=labels[index])
                elif data.spin == 2:
                    ax.plot(dx, pltband[j, :, 0], color=color[index],
                            linestyle=linestl, linewidth=linewidth, label=labels[index]+' Alpha')
                    ax.plot(dx, pltband[j, :, 1], color=color[index],
                            linestyle='--', linewidth=linewidth, label=labels[index]+' Beta')

            else:
                if data.spin == 1:
                    if isinstance(linestl, list):
                        if isinstance(linewidth, list):
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl[index], linewidth=linewidth[index])
                        else:
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl[index], linewidth=linewidth)
                    else:
                        if isinstance(linewidth, list):
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl, linewidth=linewidth[index])
                        else:
                            ax.plot(dx, pltband[j, :], color=color[index],
                                    linestyle=linestl, linewidth=linewidth)
                elif data.spin == 2:
                    ax.plot(dx, pltband[j, :, 0], color=color[index],
                            linestyle=linestl, linewidth=linewidth)
                    ax.plot(dx, pltband[j, :, 1], color=color[index],
                            linestyle='--', linewidth=linewidth)

            count1 += 1

    if (isinstance(xmin, list)) or (isinstance(xmax, list)):
        xmin = min(xmin)
        xmax = max(xmax)

    ymin = min(ymin)
    ymax = max(ymax)

    return ymin, ymax, xmin, xmax, fig, ax


def plot_compare_cry_bands(bands, energy_range, not_scaled, linestl, linewidth,
                           color, fermi, figsize, scheme, sharex, sharey,
                           fermiwidth, fermialpha):

    import sys
    import warnings

    import matplotlib.pyplot as plt
    import numpy as np

    # Error check and definition of scheme
    if scheme is None:
        n_rows = 1
        n_col = len(bands)
    else:
        if (not isinstance(scheme, tuple)) and (not isinstance(scheme, list)):
            raise ValueError('scheme needs to be a tuple or a list')
            sys.exit(1)
        elif len(scheme) > 2:
            raise ValueError(
                'scheme needs to be a tuple or a list of two elements (nrow,ncol)')
        else:
            n_rows = scheme[0]
            n_col = scheme[1]

    # Creation of the subplots
    if figsize is None:
        fig, ax = plt.subplots(nrows=n_rows, ncols=n_col,
                               sharex=sharex, sharey=sharey, figsize=figsize)
    else:
        fig, ax = plt.subplots(nrows=n_rows, ncols=n_col,
                               sharex=sharex, sharey=sharey, figsize=figsize)
    if n_rows == 1 and n_col == 1:  # When ax is not a list
        ax = [ax]
    # Scaling with different size of the same brillouin zone
    if not_scaled is False:
        reference = xmax = np.amax(bands[0].k_point_plot)
        xmin = np.amin(bands[0].k_point_plot)

    else:
        xmax = []
        xmin = []

    ymin = []
    ymax = []
    count3 = 0

    # Plot of the different band structure into the subplots
    for col in range(n_col):
        for row in range(n_rows):
            data = bands[count3]
            if count3 == 0:
                hsp = data.tick_position
            pltband = data.bands
            no_bands = np.shape(pltband)[0]
            if not_scaled is False:
                k_max = np.amax(data.k_point_plot)
                dx = (data.k_point_plot/k_max)*reference
            else:
                dx = data.k_point_plot
                xmin.append(np.amin(dx))
                xmax.append(np.amax(dx))
            ymin.append(np.amin(pltband))
            ymax.append(np.amax(pltband))
            # Effective plotting action
            for j in range(no_bands):
                if data.spin == 1:
                    if n_rows == 1:
                        ax[col].plot(
                            dx, pltband[j, :], color=color, linestyle=linestl, linewidth=linewidth)
                    else:
                        ax[row, col].plot(
                            dx, pltband[j, :], color=color, linestyle=linestl, linewidth=linewidth)
                elif data.spin == 2:
                    if n_rows == 1:
                        ax[col].plot(dx, pltband[j, :, 0], color=color,
                                     linestyle='-', linewidth=linewidth, label='Alpha')
                        ax[col].plot(dx, pltband[j, :, 0], color=color,
                                     linestyle='--', linewidth=linewidth, label='Beta')
                    else:
                        ax[row, col].plot(
                            dx, pltband[j, :, 0], color=color, linestyle='-', linewidth=linewidth, label='Alpha')
                        ax[row, col].plot(
                            dx, pltband[j, :, 0], color=color, linestyle='--', linewidth=linewidth, label='Beta')

            # Plot of the HSPs lines
            yhsp = np.linspace(np.amin(pltband)+5, np.amax(pltband)+5, 2)
            for j in hsp:
                xhsp = np.ones(2)*j
                if n_rows == 1:
                    ax[col].plot(
                        xhsp, yhsp, color='black', linewidth=0.5)
                else:
                    ax[row, col].plot(
                        xhsp, yhsp, color='black', linewidth=0.5)

            # Fermi level line plot
            xfermi = np.linspace(np.amin(pltband), np.amax(pltband), 2)
            yfermi = np.zeros(2)
            if n_rows == 1:
                ax[col].plot(
                    xfermi, yfermi, color=fermi, linewidth=fermiwidth, alpha=fermialpha)
            else:
                ax[row, col].plot(
                    xfermi, yfermi, color=fermi, linewidth=fermiwidth, alpha=fermialpha)

            # Definition of x and y limits
            if n_rows == 1:
                if sharex is not True:
                    """hsp_label = []
                    for element in k_labels:
                        if element in k_labels:
                            g = greek.get(element)
                            hsp_label.append(g)
                    ax[col].set_xticks(hsp)
                    if k_labels is not None:
                        ax[col].set_xlabels(hsp_label)"""
                    warnings.warn(
                        'The sharex = False option has not been developed yet')
                ax[col].set_xlim([np.amin(dx), np.amax(dx)])
                if (sharey is not True) and (energy_range is not None):
                    ax[col].set_ylim([energy_range[0], energy_range[1]])
                else:
                    ax[col].set_ylim([np.amin(pltband), np.amax(pltband)])
            else:
                if sharex is not True:
                    """hsp_label = []
                    for element in k_labels:
                        if element in k_labels:
                            g = greek.get(element)
                            hsp_label.append(g)
                    ax[row, col].set_xticks(hsp)
                    if k_labels is not None:
                        ax[row, col].set_xlabels(hsp_label)"""
                    warnings.warn(
                        'The sharex = False option has not been developed yet')
                ax[row, col].set_xlim([np.amin(dx), np.amax(dx)])
                if (sharey is not True) and (energy_range is not False):
                    ax[row, col].set_ylim(
                        [energy_range[0], energy_range[1]])
                else:
                    ax[row, col].set_ylim(
                        [np.amin(pltband), np.amax(pltband)])

            count3 += 1

    if (isinstance(ymin, list)) or (isinstance(ymax, list)):
        ymin = min(ymin)
        ymax = max(ymax)

    return ymin, ymax, xmin, xmax, hsp, fig, ax


def plot_cry_doss(doss, color, fermi, overlap, labels, figsize, linestl,
                  linewidth, title, beta, energy_range, dos_range, prj):
    """
    The base function to plot electron / phonon density of states.

    Args:
        doss (object): The density of states object.
        color (str or list): The color(s) of the plot(s).
        fermi (str): The color of the Fermi level line.
        overlap (bool): True if the projections should overlap, False otherwise.
        labels (str or list): The label(s) for the plot(s).
        figsize (tuple): The figure size (width, height) in inches.
        linestl (str or list): The line style(s) for the plot(s).
        linewidth (float): The width of the line(s) in points.
        title (str): The title of the plot.
        beta (str): The beta value ('up' or 'down').
        energy_range (tuple): The energy range (xmin, xmax) for the x-axis.
        dos_range (tuple): The density of states range (ymin, ymax) for the y-axis.
        prj (None or list): The projection(s) to plot.

    Returns:
        None
    """
    import sys
    import warnings
    from os import path

    import matplotlib.pyplot as plt
    import numpy as np

    # Error check on beta
    accepted_beta = ['up', 'down']
    if beta not in accepted_beta:
        raise ValueError("Beta can be defined only as: 'up' or 'down'.")

    # Plot for overlap == True
    if overlap == True:

        # Error check on color for color in overlap == true and default list of color
        if (not isinstance(color, list)) and (doss.n_proj > 1):
            warnings.warn(
                'When overlap is true color has to be a list!', stacklevel=2)
            color = ['blue', 'indigo', 'midgrey', 'crimson']
            if (len(color) < doss.n_proj) or (len(color) < len(prj)):
                raise ValueError(
                    'When overlap is true color has to be a list!')

        # Error check on labels for overlap == true
        if labels is not None:
            if not isinstance(labels, list) and (doss.n_proj > 1):
                raise ValueError(
                    'When overlap is true labels has to be a list')

        # Default for figsize
        if figsize is None:
            figsize = (12, 5)

        # Error check on color, labels, and linestl
        if (isinstance(color, list)) or (isinstance(labels, list)) or (isinstance(linestl, list)):
            if color is not None:
                if prj is None:
                    if (len(color) > doss.n_proj) or (len(color) < doss.n_proj):
                        if len(color) > doss.n_proj:
                            raise ValueError(
                                'The number of colors is greater than the number of projections!')
                        elif len(color) < doss.n_proj:
                            raise ValueError(
                                'The number of colors is less than the number of projections!')
                else:
                    if (len(color) > len(prj)) or (len(color) < len(prj)):
                        if len(color) > len(prj):
                            raise ValueError(
                                'The number of colors is greater than the number of projections required!')
                        elif len(color) < len(prj):
                            raise ValueError(
                                'The number of colors is less than the number of projections required!')

            if labels is not None:
                if prj is None:
                    if (len(labels) > doss.n_proj) or (len(labels) < doss.n_proj):
                        if len(labels) > doss.n_proj:
                            raise ValueError(
                                'The number of labels is greater than the number of projections!')
                        elif len(labels) < doss.n_proj:
                            raise ValueError(
                                'The number of labels is less than the number of projections!')
                else:
                    if (len(labels) > len(prj)) or (len(labels) < len(prj)):
                        if len(labels) > len(prj):
                            raise ValueError(
                                'The number of labels is greater than the number of projections required!')
                        elif len(color) < len(prj):
                            raise ValueError(
                                'The number of labels is less than the number of projections required!')

            if linestl is not None:
                if prj is None:
                    if (len(linestl) > doss.n_proj) or (len(linestl) < doss.n_proj):
                        if len(linestl) > doss.n_proj:
                            raise ValueError(
                                'The number of linestl is greater than the number of projections!')
                        elif len(linestl) < doss.n_proj:
                            raise ValueError(
                                'The number of linestl is less than the number of projections!')
                else:
                    if (len(linestl) > len(prj)) or (len(linestl) < len(prj)):
                        if len(linestl) > len(prj):
                            raise ValueError(
                                'The number of linestl is greater than the number of projections required!')
                        elif len(color) < len(prj):
                            raise ValueError(
                                'The number of linestl is less than the number of projections required!')

        # Creation of the figure
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        # Creation of dx for the plot
        if doss.spin == 1:
            dx = doss.doss[:, 0]
        elif doss.spin == 2:
            dx = doss.doss[:, 0, :]
            dx_alpha = doss.doss[:, 0, 0]
            dx_beta = doss.doss[:, 0, 1]

        # Determination of xmin, xmax, ymin, and ymax
        xmin = np.amin(dx)
        xmax = np.amax(dx)
        ymin = np.amin(doss.doss[1:, :, :])
        ymax = np.amax(doss.doss[1:, :, :])

        # Plot of all projections
        if prj == None:
            for projection in range(1, doss.n_proj+1):
                # for projection in range(1, 2):
                if doss.spin == 1:
                    if doss.n_proj > 1:
                        if linestl is None:
                            ax.plot(dx, doss.doss[:, projection], color=color[projection-1],
                                    label=labels[projection-1], linewidth=linewidth)
                        else:
                            ax.plot(dx, doss.doss[:, projection], color=color[projection-1],
                                    label=labels[projection-1], linestyle=linestl[projection-1], linewidth=linewidth)
                    else:
                        ax.plot(dx, doss.doss[:, projection],
                                color=color, linewidth=linewidth)
                elif doss.spin == 2:
                    if beta == accepted_beta[0]:
                        if doss.n_proj > 1:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0], color=color[projection-1],
                                    label=labels[projection-1], linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, -doss.doss[:, projection, 1], color=color[projection-1],
                                    label=labels[projection-1], linestyle='--', linewidth=linewidth)
                        else:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0],
                                    linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, -doss.doss[:, projection, 1],
                                    linestyle='--', linewidth=linewidth)
                    elif beta == accepted_beta[1]:
                        if doss.n_proj > 1:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0], color=color[projection-1],
                                    label=labels[projection-1], linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, doss.doss[:, projection, 1], color=color[projection-1],
                                    label=labels[projection-1], linestyle='--', linewidth=linewidth)

                        else:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0], color=color,
                                    linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, doss.doss[:, projection, 1], color=color,
                                    linestyle='--', linewidth=linewidth)

        # Plot of selected projections
        else:
            for index, projection in enumerate(prj):
                if doss.spin == 1:
                    if doss.n_proj > 1:
                        if linestl is None:
                            ax.plot(dx, doss.doss[:, projection], color=color[index],
                                    label=labels[index], linewidth=linewidth)
                        else:
                            ax.plot(dx, doss.doss[:, projection], color=color[index],
                                    label=labels[index], linestyle=linestl[index], linewidth=linewidth)
                    else:
                        ax.plot(dx, doss.doss[:, projection],
                                color=color, linewidth=linewidth)
                elif doss.spin == 2:
                    if beta == accepted_beta[0]:
                        if doss.n_proj > 1:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0], color=color[index],
                                    label=labels[index], linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, -doss.doss[:, projection, 1], color=color[index],
                                    label=labels[index], linestyle='--', linewidth=linewidth)
                        else:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0],
                                    linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, -doss.doss[:, projection, 1],
                                    linestyle='--', linewidth=linewidth)
                    elif beta == accepted_beta[1]:
                        if doss.n_proj > 1:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0], color=color[index],
                                    label=labels[index], linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, doss.doss[:, projection, 1], color=color[index],
                                    label=labels[index], linestyle='--', linewidth=linewidth)
                        else:
                            ax.plot(dx_alpha, doss.doss[:, projection, 0], color=color,
                                    linestyle='-', linewidth=linewidth)
                            ax.plot(dx_beta, doss.doss[:, projection, 1], color=color,
                                    linestyle='--', linewidth=linewidth)

        yfermi_level = np.linspace(ymin*1.05, ymax*1.05, 2)
        xfermi_level = np.zeros(2)
        if beta == accepted_beta[0]:
            ax.plot(xfermi_level, yfermi_level, color=fermi, linewidth=1.5)
        else:
            yfermi_level = np.linspace(ymin*1.05, ymax*1.05, 2)
            ax.plot(xfermi_level, yfermi_level, color=fermi, linewidth=1.5)
        y_zero = np.zeros(2)
        x_zero = np.linspace(xmin, xmax, 2)
        ax.plot(x_zero, y_zero, color='black', linewidth=0.4)
        if dos_range is not None:
            ymin = dos_range[0]
            ymax = dos_range[1]
            plt.ylim(ymin, ymax)
        else:
            if doss.spin == 1:
                ymin = 0
                plt.ylim(ymin, ymax*1.05)
            elif doss.spin == 2:
                if beta == accepted_beta[0]:
                    ymin = 0
                    plt.ylim(ymin, ymax*1.05)
                elif beta == accepted_beta[1]:
                    plt.ylim(ymin*1.05, ymax*1.05)

    # Plot for overlap == False
    else:
        # Error checks on colors, labels, and linestl
        if isinstance(color, list):
            warnings.warn('When overlap is false color should be a string!',
                          stacklevel=2)
            color = 'blue'

        if isinstance(labels, list):
            warnings.warn('When overlap is false labels should be a string!',
                          stacklevel=2)
            labels = None

        if isinstance(linestl, list):
            warnings.warn('When overlap is false color should be a string!',
                          stacklevel=2)
            linestl = '-'

        # Creation of dx for the plot
        if doss.spin == 1:
            dx = doss.doss[:, 0]
        elif doss.spin == 2:
            dx = doss.doss[:, 0, :]
            dx_alpha = doss.doss[:, 0, 0]
            dx_beta = doss.doss[:, 0, 1]

        # Determination of xmin and xmax
        xmin = np.amin(dx)
        xmax = np.amax(dx)

        # Creation xfermi, x_zero, and y_zero
        xfermi = np.zeros(2)
        x_zero = np.linspace(xmin, xmax, 2)
        y_zero = np.zeros(2)

        # Creation of subplots for all projections
        if prj is None:
            fig, ax = plt.subplots(
                nrows=doss.n_proj, ncols=1, sharex=True, figsize=figsize)
            # If only one projection is generated, ax is not subscriptable
            if doss.n_proj == 1:
                ax = [ax]
        # Creation of subplots for selected projections
        else:
            fig, ax = plt.subplots(
                nrows=len(prj), ncols=1, sharex=True, figsize=figsize)
            # If only one projection is generated, ax is not subscriptable
            if len(prj) == 1:
                ax = [ax]

        # Plot for all projections
        if prj is None:
            for projection in range(doss.n_proj):
                if doss.spin == 1:
                    ymin = 0
                    ymax = np.amax(doss.doss[:, projection+1])
                    yfermi = np.linspace(ymin*1.05, ymax*1.05, 2)
                    ax[projection].plot(dx, doss.doss[:, projection+1],
                                        color=color, linestyle=linestl, linewidth=linewidth)
                    ax[projection].plot(
                        xfermi, yfermi, color=fermi, linewidth=1.5)
                    if dos_range is not None:
                        ymin = dos_range[0]
                        ymax = dos_range[1]
                    ax[projection].set_ylim(ymin, ymax)
                elif doss.spin == 2:
                    if beta == accepted_beta[0]:
                        ymin = 0
                        ymax = np.amax(doss.doss[:, projection+1, :])
                        yfermi = np.linspace(ymin, ymax, 2)
                        ax[projection].plot(dx_alpha, doss.doss[:, projection+1, 0], color=color,
                                            linestyle='-', linewidth=linewidth, label='Alpha')
                        ax[projection].plot(dx_beta, -doss.doss[:, projection+1, 1], color=color,
                                            linestyle='--', linewidth=linewidth, label='Beta')
                        ax[projection].plot(
                            xfermi, yfermi, color=fermi, linewidth=1.5)
                        if dos_range is not None:
                            ymin = dos_range[0]
                            ymax = dos_range[1]
                        ax[projection].set_ylim(ymin, ymax)
                    elif beta == accepted_beta[1]:
                        ymin = -np.amax(doss.doss[:, projection+1, :])
                        ymax = np.amax(doss.doss[:, projection+1, :])
                        yfermi = np.linspace(ymin, ymax, 2)
                        ax[projection].plot(dx_alpha, doss.doss[:, projection+1, 0], color=color,
                                            linestyle='-', linewidth=linewidth, label='Alpha')
                        ax[projection].plot(dx_beta, doss.doss[:, projection+1, 1], color=color,
                                            linestyle='--', linewidth=linewidth, label='Beta')
                        ax[projection].plot(x_zero, y_zero, linewidth=0.4)
                        ax[projection].plot(
                            xfermi, yfermi, color=fermi, linewidth=2.5)
                        if dos_range is not None:
                            ymin = dos_range[0]
                            ymax = dos_range[1]
                        ax[projection].set_ylim(ymin, ymax)

        # Plot for selected projections
        else:
            for index, projection in enumerate(prj):
                if doss.spin == 1:
                    ymin = 0
                    ymax = np.amax(doss.doss[:, projection])
                    yfermi = np.linspace(ymin*1.05, ymax*1.05, 2)
                    ax[index].plot(dx, doss.doss[:, projection],
                                   color=color, linestyle=linestl, linewidth=linewidth)
                    ax[index].plot(xfermi, yfermi, color=fermi, linewidth=1.5)
                    if dos_range is not None:
                        ymin = dos_range[0]
                        ymax = dos_range[1]
                    ax[index].set_ylim(ymin, ymax)
                elif doss.spin == 2:
                    if beta == accepted_beta[0]:
                        ymin = 0
                        ymax = np.amax(doss.doss[:, projection, :])
                        yfermi = np.linspace(ymin*1.05, ymax*1.05, 2)
                        ax[index].plot(dx_alpha, doss.doss[:, projection, 0], color=color,
                                       linestyle='-', linewidth=linewidth, label='Alpha')
                        ax[index].plot(dx_beta, -doss.doss[:, projection, 1], color=color,
                                       linestyle='--', linewidth=linewidth, label='Beta')
                        ax[index].plot(
                            xfermi, yfermi, color=fermi, linewidth=1.5)
                        if dos_range is not None:
                            ymin = dos_range[0]
                            ymax = dos_range[1]
                        ax[index].set_ylim(ymin, ymax)
                    elif beta == accepted_beta[1]:
                        ymin = -np.amax(doss.doss[:, projection, :])
                        ymax = np.amax(doss.doss[:, projection, :])
                        yfermi = np.linspace(ymin*1.05, ymax*1.05, 2)
                        ax[index].plot(dx_alpha, doss.doss[:, projection, 0], color=color,
                                       linestyle='-', linewidth=linewidth, label='Alpha')
                        ax[index].plot(dx_beta, doss.doss[:, projection, 1], color=color,
                                       linestyle='--', linewidth=linewidth, label='Beta')
                        ax[index].plot(x_zero, y_zero, linewidth=0.4)
                        ax[index].plot(
                            xfermi, yfermi, color=fermi, linewidth=1.5)
                        if dos_range is not None:
                            ymin = dos_range[0]
                            ymax = dos_range[1]
                        ax[index].set_ylim(ymin, ymax)

        if (doss.spin == 2) and (beta == accepted_beta[1]):
            plt.legend()

    if energy_range is not None:
        xmin = energy_range[0]
        xmax = energy_range[1]
    plt.xlim(xmin, xmax)

    if title is not None:
        fig.suptitle(title)
    if labels is not None:
        plt.legend()

    return fig, ax


def plot_cry_es(bands, doss, k_labels, color_bd, color_doss, fermi, energy_range, linestl_bd,
                linestl_doss, linewidth, prj, figsize, labels, dos_range, title, dos_beta, legend):
    """
    The base function to plot electron / phonon band structure + DOS

    Args:
        bands (object): Object containing band structure data
        doss (object): Object containing density of states data
        k_labels (list or None): List of labels for high symmetry points along the path
        color_bd (str): Color for the band structure plot
        color_doss (str or list or tuple): Color(s) for the density of states plot
        fermi (str): Color for the Fermi level lines
        energy_range (list or None): Range of energy values for the y-axis
        linestl_bd (str): Linestyle for the band structure plot
        linestl_doss (str or list or tuple or None): Linestyle(s) for the density of states plot
        linewidth (float): Width of the lines
        prj (list or None): List of projection indices for plotting specific projections
        figsize (tuple): Figure size (width, height)
        labels (str or list or tuple or None): Labels for the density of states plot
        dos_range (list or None): Range of the density of states plot
        title (str or None): Title of the figure
        dos_beta (str): Beta state for the density of states plot ('up' or 'down')
        legend (bool): Enables or disables the legend

    Returns:
        fig (object): Figure object containing the plotted data
    """
    import warnings
    from os import path

    import matplotlib.pyplot as plt
    import numpy as np

    # Dictionary of greek letters for the band plot in the electronic structure
    greek = {'Alpha': '\u0391', 'Beta': '\u0392', 'Gamma': '\u0393', 'Delta': '\u0394', 'Epsilon': '\u0395', 'Zeta': '\u0396', 'Eta': '\u0397',
             'Theta': '\u0398', 'Iota': '\u0399', 'Kappa': '\u039A', 'Lambda': '\u039B', 'Mu': '\u039C', 'Nu': '\u039D', 'Csi': '\u039E',
             'Omicron': '\u039F', 'Pi': '\u03A0', 'Rho': '\u03A1', 'Sigma': '\u03A3', 'Tau': '\u03A4', 'Upsilon': '\u03A5', 'Phi': '\u03A6',
             'Chi': '\u03A7', 'Psi': '\u03A8', 'Omega': '\u03A9', 'Sigma_1': '\u03A3\u2081'}

    if isinstance(color_doss, list) or isinstance(color_doss, tuple):
        # Error check on linestl_doss length when the prj kwargs is not used
        if (prj is None) and (len(color_doss) != doss.n_proj):
            if len(color_doss) > doss.n_proj:
                warnings.warn('You have a number of linestl_doss elements is greater than the number of projection!',
                              stacklevel=2)
            else:
                raise ValueError(
                    "You don't have enough elements in linestl_doss for the number of projection required")
        # Error check on linestl_doss length when the prj kwargs is used
        elif (prj is not None) and (len(color_doss) != len(prj)):
            if len(color_doss) > len(prj):
                warnings.warn('You have a number of linestl_doss element greater than the number of projection required(prj elements)!',
                              stacklevel=2)
            else:
                raise ValueError(
                    "You don't have enough elements in linestl_doss for the number of projection required(prj elements)")
    else:
        if prj == None:
            nprj = doss.n_proj
        else:
            nprj = len(prj)
        if nprj > 1:
            warnings.warn(
                'Only one color / no color is given. All DOS projections are in the same color.')
        color_doss = [color_doss for i in range(nprj)]

    if isinstance(labels, list) or isinstance(labels, tuple):
        # Error check on labels length when the prj kwargs is not used
        if (prj is None) and (len(labels) != doss.n_proj):
            if len(labels) > doss.n_proj:
                warnings.warn(
                    'You have a number of labels greater than the number of projection!')
            else:
                raise ValueError(
                    "You don't have enough elements in labels for the numeber of projection required")
        # Error check on labels length when the prj kwargs is used
        elif (prj is not None) and (len(labels) != len(prj)):
            if len(labels) > len(prj):
                warnings.warn('You have a number of linestl_doss element greater than the number of projection required(prj elements)!',
                              stacklevel=2)
            else:
                raise ValueError(
                    "You don't have enough elements in linestl_doss for the number of projection required(prj elements)")
    else:
        if prj == None:
            nprj = doss.n_proj
        else:
            nprj = len(prj)
        if nprj > 1:
            warnings.warn('Label not given. No label is available for DOS.')
        labels = [None for i in range(nprj)]

    # DOS beta state
    if dos_beta == 'up':
        spin_idx = -1  # DOS in crystal output is distinguished by +/- sign already
        line_0 = False
    elif dos_beta == 'down':
        spin_idx = 1
        line_0 = True
    else:
        raise ValueError("'dos_beta' should be either 'up' or 'down'.")
    if doss.doss.shape[2] == 2:
        doss.doss[:, :, 1] = doss.doss[:, :, 1] * spin_idx

    # Definition and creation of the figure and the axes
    fig, ax = plt.subplots(nrows=1, ncols=2, gridspec_kw={'width_ratios': [2, 1]},
                           sharex=False, sharey=True, figsize=figsize)
    if title != None:
        fig.suptitle(title)

    # Definition of the hsp position variables
    hsp = bands.tick_position

    # Error check on k_labels lenght against the HSP poisitions
    if k_labels is not None:
        if len(hsp) != len(k_labels):
            if len(hsp) > len(k_labels):
                raise ValueError(
                    'You have specified a number of label smaller than the number of High Simmetry Point along the path')
            elif len(hsp) < len(k_labels):
                raise ValueError(
                    'You have more labels than the High Simmetry point along the path')

    # Local variable definition for the band plot
    dx_bd = bands.k_point_plot
    pltband = bands.bands
    no_bands = np.shape(pltband)[0]
    ymin_bd = np.amin(pltband)
    ymax_bd = np.amax(pltband)
    xmin_bd = np.amin(dx_bd)
    xmax_bd = np.amax(dx_bd)
    count1 = 0
    count2 = 0

    # band plot
    for i in range(no_bands):
        if bands.spin == 1:
            ax[0].plot(dx_bd, pltband[i, :], color=color_bd,
                       linestyle=linestl_bd, linewidth=linewidth)

        elif bands.spin == 2:
            if count1 == count2:
                ax[0].plot(dx_bd, pltband[i, :, 0], color=color_bd,
                           linestyle='-', linewidth=linewidth, label='Alpha')
                ax[0].plot(dx_bd, pltband[i, :, 1], color=color_bd,
                           linestyle='--', linewidth=linewidth, label='Beta')
            else:
                ax[0].plot(dx_bd, pltband[i, :, 0], color=color_bd,
                           linestyle='-', linewidth=linewidth)
                ax[0].plot(dx_bd, pltband[i, :, 1], color=color_bd,
                           linestyle='--', linewidth=linewidth)

        count1 += 1

    # Definition of dx for the doss plot
    if doss.spin == 1:
        dx_dos = doss.doss[:, 0]
    elif doss.spin == 2:
        dx_dos = doss.doss[:, 0, :]
        dx_alpha = doss.doss[:, 0, 0]
        dx_beta = doss.doss[:, 0, 1]

    # Determination of xmin, xmax, ymin, and ymax
    if prj != None:
        argplt = prj
    else:
        argplt = range(1, doss.doss.shape[1])

    # Plot a vertical line at 0 DOS. Only for spin polarized cases
    if line_0 == True and doss.spin == 2:
        xmin_dos = np.amin(doss.doss[:, argplt, 1])
        xmax_dos = np.amax(doss.doss[:, argplt, 0])
        # make the scale symmetric, for comparison
        xmin_dos = -max([abs(xmin_dos), abs(xmax_dos)])
        xmax_dos = -xmin_dos
    else:
        xmin_dos = 0.
        xmax_dos = np.amax(doss.doss[:, argplt, :])

    ymin_dos = np.amin(dx_dos)
    ymax_dos = np.amax(dx_dos)

    # Plot of all projections
    if prj is None:
        for projection in range(1, doss.n_proj+1):
            if doss.spin == 1:

                if doss.n_proj > 1:

                    if linestl_doss is None:
                        ax[1].plot(doss.doss[:, projection], dx_dos, color=color_doss[projection-1],
                                   label=labels[projection-1], linewidth=linewidth)

                    else:
                        ax[1].plot(doss.doss[:, projection], dx_dos, color=color_doss[projection-1],
                                   label=labels[projection-1], linestyle=linestl_doss[projection-1], linewidth=linewidth)

                else:
                    ax[1].plot(doss.doss[:, projection], dx_dos, color=color_doss[0],
                               linewidth=linewidth)

            elif doss.spin == 2:
                if doss.n_proj > 1:
                    ax[1].plot(doss.doss[:, projection, 0], dx_alpha, color=color_doss[projection-1],
                               label=labels[projection-1], linestyle='-', linewidth=linewidth)
                    ax[1].plot(doss.doss[:, projection, 1], dx_beta, color=color_doss[projection-1],
                               label=labels[projection-1], linestyle='--', linewidth=linewidth)
                else:
                    ax[1].plot(doss.doss[:, projection, 0], dx_alpha, color=color_doss[0],
                               linestyle='-', linewidth=linewidth)
                    ax[1].plot(doss.doss[:, projection, 1], dx_beta, color=color_doss[0],
                               linestyle='--', linewidth=linewidth)

    # Plot of a selected number of projections
    else:
        for index, projection in enumerate(prj):
            if doss.spin == 1:
                if doss.n_proj > 1:
                    if linestl_doss is None:
                        ax[1].plot(doss.doss[:, projection], dx_dos, color=color_doss[index],
                                   label=labels[index], linewidth=linewidth)
                    else:
                        ax[1].plot(doss.doss[:, projection], dx_dos, color=color_doss[index],
                                   label=labels[index], linestyle=linestl_doss[index], linewidth=linewidth)
                else:
                    ax[1].plot(doss.doss[:, projection], dx_dos, color=color_doss[0],
                               linewidth=linewidth)
            elif doss.spin == 2:
                if doss.n_proj > 1:
                    ax[1].plot(doss.doss[:, projection, 0], dx_alpha, color=color_doss[index],
                               label=labels[index], linestyle='-', linewidth=linewidth)
                    ax[1].plot(doss.doss[:, projection, 1], dx_beta, color=color_doss[index],
                               label=labels[index], linestyle='--', linewidth=linewidth)
                else:
                    ax[1].plot(doss.doss[:, projection, 0], dx_alpha, color=color_doss[0],
                               linestyle='-', linewidth=linewidth)
                    ax[1].plot(doss.doss[:, projection, 1], dx_beta, color=color_doss[0],
                               linestyle='--', linewidth=linewidth)

    if ymin_bd > ymin_dos:
        ymin = ymin_dos
    elif ymin_bd <= ymin_dos:
        ymin = ymin_bd

    if ymax_bd >= ymax_dos:
        ymax = ymax_bd
    elif ymax_bd < ymax_dos:
        ymax = ymax_dos

    xmax_bd = hsp[len(hsp)-1]
    # Plot of HSP lines
    yhsp = np.linspace(ymin-5, ymax+5, 2)
    for j in hsp:
        xhsp = np.ones(2)*j
        ax[0].plot(xhsp, yhsp, color='black', linewidth=0.5)

    # Creation if HSP label ticks
    hsp_label = []
    if k_labels is not None:
        for n in k_labels:
            if n in greek:
                g = greek.get(n)
                hsp_label.append(g)
            else:
                hsp_label.append(n)

    ax[0].set_xticks(hsp)
    if k_labels is not None:
        ax[0].set_xticklabels(hsp_label)

    xfermi_bd = np.linspace(xmin_bd, xmax_bd, 2)
    xfermi_dos = np.linspace(xmin_dos*1.05, xmax_dos*1.05, 2)
    yfermi = np.zeros(2)

    # Plot of fermi level lines both in the band and the doss plot
    ax[0].plot(xfermi_bd, yfermi, color=fermi, linewidth=1.5)
    ax[1].plot(xfermi_dos, yfermi, color=fermi, linewidth=1.5)

    ax[0].set_xlim(xmin_bd, xmax_bd)

    if dos_range is not None:
        xmin_dos = dos_range[0]
        xmax_dos = dos_range[1]

    # if (prj is None) and (doss.n_proj not in prj):
    #    xmax_dos = np.amax(doss.doss[:, 1:doss.n_proj-1, :])

    if line_0 == True:
        ax[1].plot(np.zeros([2,]), [ymin, ymax], color='black', linewidth=0.5)
    ax[1].set_xlim(xmin_dos*1.05, xmax_dos*1.05)

    if energy_range is not None:
        ymin = energy_range[0]
        ymax = energy_range[1]

    plt.ylim(ymin, ymax)
    if legend:
        plt.legend()

    return fig, ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""

import numpy as np
from scipy import constants


def H_to_eV(energy):
    # Conversion from Hartree to eV
    return energy/constants.physical_constants['electron volt-hartree relationship'][0]


def eV_to_H(energy):
    # Conversion from eV to Hartree
    return energy*constants.physical_constants['electron volt-hartree relationship'][0]


def H_to_kjmol(energy):
    # Conversion from Hartree to kJ / mol
    return energy*(constants.physical_constants['Hartree energy'][0] * 1e-3 * constants.Avogadro)


def kjmol_to_H(energy):
    # Conversion from kJ / mol to Hartree
    return energy/(constants.physical_constants['Hartree energy'][0] * 1e-3 * constants.Avogadro)


def au_to_angstrom(length):
    return length*(constants.physical_constants['atomic unit of length'][0] * 1e10)


def angstrom_to_au(length):
    return length/(constants.physical_constants['atomic unit of length'][0] * 1e10)


def cm_to_thz(freq):
    # Conversion from cm^-1 to THz
    return freq*(constants.physical_constants['speed of light in vacuum'][0] * 1e-10)


def thz_to_cm(freq):
    # Conversion from cm^-1 to THz
    return freq/(constants.physical_constants['speed of light in vacuum'][0] * 1e-10)


def hartree_to_thz(freq):
    # Conversion from frequency in AU(angular) to THz(linear)
    return freq / (1e12 * constants.physical_constants['atomic mass unit-hartree relationship'][0] / constants.physical_constants['atomic mass unit-hertz relationship'][0])


def thz_to_hartree(freq):
    # Conversion from frequency in THz(linear) to AU(angular)
    return freq * (1e12 * constants.physical_constants['atomic mass unit-hartree relationship'][0] / constants.physical_constants['atomic mass unit-hertz relationship'][0])


def amu_to_me(mass):
    # Conversion from unified atomic mass unit to electron mass (mass unit in AU)
    return mass * (constants.physical_constants['unified atomic mass unit'][0] / constants.m_e)


def me_to_amu(mass):
    # Conversion from electron mass (mass unit in AU) to unified atomic mass unit
    return mass / (constants.physical_constants['unified atomic mass unit'][0] / constants.m_e)
